"""

Kubernetes workload plugin for creating infra from a multi-doc yaml file

"""

import logging
from typing import List, Any, Dict
import yaml

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

logger = logging.getLogger('metta.contrib.kubernetes.workload.yaml')

KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL = 'kubernetes'
KUBERNETES_YAML_WORKLOAD_CONFIG_BASE = 'workload.yaml'

KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE = "default"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE = "file"
KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML = "yaml"


class KubernetesYamlWorkloadPlugin(WorkloadBase):
    """ Kubernetes workload class """

    def __init__(self, environment, instance_id,
                 label: str = KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_YAML_WORKLOAD_CONFIG_BASE):
        """ Run the super constructor but also set class properties

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------

        config_file (str): String path to the kubernetes config file to use

        """
        WorkloadBase.__init__(self, environment, instance_id)

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """ Create a workload instance from a set of fixtures

        Parameters:
        -----------

        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """

        try:
            client = fixtures.get_plugin(
                type=Type.CLIENT, plugin_id='metta_kubernetes')
        except KeyError as e:
            raise NotImplementedError(
                "Workload could not find the needed client: {}".format('metta_kubernetes'))

        workload_config = self.environment.config.load(self.config_label)

        namespace = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE], default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE)

        """ YAML config needs to come from either a yaml file path or inline yaml config """
        yaml = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML], default=[])
        file = workload_config.get(
            [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE], default=None)

        if yaml is None and file is None:
            raise ValueError(
                "Either inline yaml or a file path to a yaml is required.")

        return KubernetesYamlWorkloadInstance(
            client, namespace, data=yaml, file=file)

    def info(self, deep: bool = False):
        """ Return dict data about this plugin for introspection """
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'namespace': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_NAMESPACE], default=KUBERNETES_YAML_WORKLOAD_CONFIG_DEFAULT_NAMESPACE),
                    'yaml': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_YAML], default='None'),
                    'file': workload_config.get(
                        [self.config_base, KUBERNETES_YAML_WORKLOAD_CONFIG_KEY_FILE], default='None')
                },
                'required_fixtures': {
                    'kubernetes': {
                        'type': Type.CLIENT.value,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


class KubernetesYamlWorkloadInstance(WorkloadInstanceBase):

    def __init__(self, client, namespace: str,
                 data: Dict[str, Any], file: str):
        self.client = client
        self.namespace = namespace
        self.data = data
        self.file = file

        self.k8s_objects = []

    def apply(self):
        """ Run the workload """
        if self.file:
            with open(self.file) as r:
                resources_yaml = yaml.safe_load_all(r)

                for resource in resources_yaml:
                    self.k8s_objects.append(self.client.utils_create_from_dict(
                        data=resource,
                        namespace=self.namespace
                    ))

        else:
            self.k8s_object.append(self.client.utils_create_from_dict(
                data=self.data,
                namespace=self.namespace
            ))
        return self.k8s_objects



"""

Test the binhelper utility plugin

"""
import unittest
import tempfile
import os
import shutil
import logging

from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

from mirantis.testing.metta import new_environment, environment_names, get_environment
from mirantis.testing.metta.environment import Environment

logger = logging.getLogger('test-binhelper')


CONFIG_DATA = {
    'fixtures': {
        'bin-helper': {
            'type': 'metta.plugins.utility',
            'plugin_id': 'bin-helper',

            'from_config': True,

            'path': {
                'local': "THIS WILL GET SET IN setUpClass",
                'add_to_path': True,
            },
            'platforms': {
                'Linux-x86_64': {
                    'kubectl': {
                        'url': 'https://dl.k8s.io/release/v1.20.4/bin/linux/amd64/kubectl',
                        'version': 'v1.20.4',
                    },
                    'launchpad': {
                        'url': 'https://github.com/Mirantis/launchpad/releases/download/1.2.0-rc.1/launchpad-linux-x64',
                        'version': '1.2.0-rc.1',
                        'copy': {
                            'launchpad': 'launchpad-linux-x64'
                        }
                    },
                    'helm': {
                        'url': 'https://get.helm.sh/helm-v3.5.3-linux-amd64.tar.gz',
                        'version': 'v3.5.3',
                        'copy': {
                            'helm': 'linux-amd64/helm'
                        }
                    },
                    'terraform': {
                        'url': 'https://releases.hashicorp.com/terraform/0.14.7/terraform_0.14.7_linux_amd64.zip',
                        'version': '0.14.7',
                        'copy': {
                            'terraform': 'terraform'
                        }
                    },
                    'sonobuoy': {
                        'url': 'https://github.com/vmware-tanzu/sonobuoy/releases/download/v0.50.0/sonobuoy_0.50.0_linux_amd64.tar.gz',
                        'version': 'v0.50.0',
                        'copy': {
                            'sonobuoy': 'sonobuoy'
                        }
                    }
                }
            }
        }
    }
}
""" Config data used to create a binhelper fixture

We will use a tempdir for the local path, which will be assigned in the
constructor so that we can manage the context.

"""


""" TESTS """


class BinHelperTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """
            We do 2 important things here:

            1. empty out the ENV of a PATH variable so that binhelper doesn't
               detect any bins in the test system
            2. load the test config but overwrite the local path with a tempdir
               for proper testing.  We will rmtree the tempdir in the desctructor

        """
        os.environ['PATH'] = ''
        cls.temp_bin_dir = tempfile.mkdtemp(prefix='metta_test_')
        cls.config = CONFIG_DATA
        cls.config['fixtures']['bin-helper']['path']['local'] = cls.temp_bin_dir

    @classmethod
    def tearDownClass(cls):
        """ delete the temp dir created in the constructor """
        shutil.rmtree(cls.temp_bin_dir)

    def _environment(self) -> Environment:
        """ Create an environment object, add the common config source data """
        if not 'default' in environment_names():

            environment = new_environment(name='default')
            environment.config.add_source(
                PLUGIN_ID_SOURCE_DICT).set_data(self.config)

            # this looks magical, but it works because we have structured the
            # fixtures DICT to match default values
            environment.add_fixtures_from_config()

        return get_environment(name='default')

    """ Tests """

    def test_config_basics(self):
        """ some sanity testing on the shared config object """
        logger.warning(
            "Creating environment object, which will try to download all of the bins")
        environment = self._environment()

        fixtures_config = environment.config.load('fixtures')
        self.assertEqual(fixtures_config.get(
            'bin-helper.plugin_id'), "bin-helper")
        self.assertEqual(
            fixtures_config.get('bin-helper.path.local'),
            self.temp_bin_dir)

        self.assertTrue(self.temp_bin_dir in os.environ['PATH'])

        self.assertEqual(len(fixtures_config.get(
            'bin-helper.platforms').keys()), 1)
        self.assertEqual(len(fixtures_config.get(
            'bin-helper.platforms.Linux-x86_64').keys()), 5)

        print('Path: {}'.format(self.temp_bin_dir))

    def test_predefined_bins(self):
        """ test that the binhelper automatically loaded a bin from config """

        self.assertIsNotNone(shutil.which('kubectl'))
        self.assertIsNotNone(shutil.which('launchpad'))
        self.assertIsNotNone(shutil.which('terraform'))
        self.assertIsNotNone(shutil.which('helm'))
        self.assertIsNotNone(shutil.which('sonobuoy'))

        list_dir = os.listdir(self.temp_bin_dir)
        self.assertTrue('kubectl' in list_dir)
        self.assertTrue('launchpad' in list_dir)
        self.assertTrue('terraform' in list_dir)
        self.assertTrue('helm' in list_dir)
        self.assertTrue('sonobuoy' in list_dir)




import logging
from typing import List

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import ClientBase

from .launchpad import LaunchpadClient

logger = logging.getLogger('metta.contrib.kubernetes.client')


METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID = "metta_launchpad_exec"


class LaunchpadExecClientPlugin(ClientBase):
    """ Client for exec into hosts using launchpad """

    def __init__(self, environment: Environment,
                 instance_id: str, client: LaunchpadClient):
        """

        Parameters:
        -----------

        client (LaunchpadClient) : A configured launchpad client to use to run
            exec commands

        """
        ClientBase.__init__(self, environment, instance_id)

        self.client = client

    def hosts(self, deep: bool = False):
        """ list the hosts in the cluster """

        config = self.client.describe_config()

        if deep:
            list = [host for host in config['spec']['hosts']]
        else:
            list = []
            for host in config['spec']['hosts']:
                list_host = {
                    'role': host['role']
                }
                if 'ssh' in host:
                    list_host.update({
                        'is_windows': False,
                        'address': host['ssh']['address']
                    })
                if 'winrm' in host:
                    list_host.update({
                        'is_windows': True,
                        'address': host['winrm']['address']
                    })

                list.append(list_host)

        return list

    def exec(self, cmds: List[str], host_index: int):
        """ Exec a string arg list command on a single host """

        client_config = self.client.describe_config()
        hosts = client_config['spec']['hosts']
        host = hosts[host_index]

        return self.client.exec(host_index, cmds)



import logging
from typing import Dict

from configerus.loaded import Loaded, LOADED_KEY_ROOT
from mirantis.testing.metta.output import OutputBase

logger = logging.getLogger('metta.contrib.common.output.dict')


class DictOutputPlugin(OutputBase):
    """ metta Output plugin a Dict output type

    This output plugin leverages configurators features, treating the dict as
    a config source, in order to get validation and navigations.

    """

    def __init__(self, environment, instance_id,
                 data: Dict = {}, validator: str = ''):
        """ Run the super constructor but also set class properties

        Here we treat the data dict as a configerus.loaded.Loaded instance,
        with out config as a parent, so that we can leverage the configerus
        tools for searching, formating and validation.

        a configerus.loaded.Loaded object is kept for the config.

        Parameters:
        -----------

        data (Dict) : any dict data to be stored

        validator (str) : a configerus validator target if you want valdiation
            applied to the data before it is added.

        Raises:
        -------

        A configerus.validate.ValidationError is raised if you passed in a
        validator target and validation failed.

        An AssertionError is raised if you didn't pass in a Dict.

        """
        super(OutputBase, self).__init__(environment, instance_id)

        self.set_data(data, validator)

    def set_data(self, data: Dict = {}, validator: str = ''):
        """ Re-set the data for the output """
        assert isinstance(
            data, dict), "Expected Dict of data, got {}".format(data)

        if validator:
            self.environment.config.validate(data, validator)

        mock_instance_id = 'dict-output-{}'.format(self.instance_id)
        self.loaded = Loaded(
            data=data,
            parent=self.environment.config,
            instance_id=mock_instance_id)

    def get_output(self, key: str = LOADED_KEY_ROOT, validator: str = ''):
        """ retrieve an output

        Because we treated that data as a high-priority configerus source with
        a custom label, we can retrieve data from that source easily and also
        leverage other configerus options such as templating and validation

        If you don't pass a key, you will get the entire data value

        Parameters:
        -----------

        key (str) : you can optionally pass a key to retrieve only a part of the
            data structure.  This uses the configerus .get() command which uses
            dot "." notation to descend a tree.

        validator (str) : you can tell configerus to apply a validator to the
            return value

        Returns:
        --------

        Any of retreived data in the assigned data, as though you were making
        a configerus.loaded.Loaded.get()

        Raises:
        -------

        AttributeError if you are trying to get output before you have
        assigned data using .arguments()

        configerus.validate.ValidationError if you passed in a validator and
        validation failed.

        """
        return self.loaded.get(key, validator=validator)

    def info(self):
        """ Return dict data about this plugin for introspection """
        return {
            'output': {
                'data': self.loaded.data
            }
        }



import os
import logging
import getpass
from datetime import datetime
import platform

import appdirs

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH as CONFIGERUS_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT as CONFIGERUS_SOURCE_DICT
from mirantis.testing.metta.environment import Environment
import mirantis.testing.metta as metta

logger = logging.getLogger("metta.common")

DIR = os.path.abspath(os.getcwd())

METTA_COMMON_APP_NAME = "metta"
""" used for some path building for config sources """
METTA_COMMON_CONFIG_USER_INSTANCE_ID = "user"
""" config source instance id for user path """
METTA_COMMON_CONFIG_PROJECT_DYNAMIC_INSTANCE_ID = "project_dynamic"
""" config source instance id for dynamic dict of commonly used value """
METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS = 35
""" Config source priority for added common config """
METTA_COMMON_PROJECT_CONFIG_SUBPATH = "config"
""" convention for configuration path as a subfolder of the cwd path """


def add_common_config(environment: Environment):
    """ Add some common configuration sources

    The following could be added:

    1. a `${PWD}/config` path source named "project_config" if that path exists
    2. a user defaults `~/metta` path source if that path exists
    3. a dynamic dict of common values:
        i. `user:id` pulled from the os
        ii. `global:datetime` to provide a constant datetime that can be used
           for consistent labelling
        iii `path:project` path to ${PWD} in case you want to build file paths
           based on that path

        @note we will likely grow this with common sense values

    Parameters:
    -----------

    environment (Environment) : METTA environment to be modified

    """

    project_root_path = find_project_root_path()

    # a user config path (like ~/.config/metta) may contain config
    user_conf_path = appdirs.user_config_dir(METTA_COMMON_APP_NAME)
    if not environment.config.has_source(
            METTA_COMMON_CONFIG_USER_INSTANCE_ID) and os.path.isdir(user_conf_path):
        environment.config.add_source(
            CONFIGERUS_SOURCE_PATH,
            METTA_COMMON_CONFIG_USER_INSTANCE_ID,
            METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS).set_path(user_conf_path)

    # Add some dymanic values for config
    environment.config.add_source(CONFIGERUS_SOURCE_DICT, METTA_COMMON_CONFIG_PROJECT_DYNAMIC_INSTANCE_ID, priority=METTA_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS).set_data({
        "user": {
            "id": getpass.getuser()  # override user id with a host value
        },
        'environment': {
            'name': environment.name
        },
        "global": {
            "datetime": datetime.now(),  # use a single datetime across all checks
        },
        environment.config.paths_label(): {  # special config label for file paths, usually just "paths"
            # you can use "paths:project" in config to substitute this path
            "project": project_root_path
        },
        'platform': {
            'machine': platform.machine(),
            'system': platform.system(),
            'processor': platform.processor(),
            'version': platform.version(),
            'uname': platform.uname()
        }
    })


def find_project_root_path():
    """ try to find a project root path

    We start looking in the cwd for certain marker files, and if we don't find
    any then we check the parent, recursively.

    If we never find a marker file then we assume that the current path is the
    root.

    """

    MARKER_FILES = {
        'metta.py',
        'mettac.py',
        'conftest.py',
        'pytest.ini',
    }

    # Try to add a path from cwd and up that contains a mettac.py file
    check_path = DIR
    while check_path:
        if check_path == '/':
            return DIR

        for MARKER_FILE in MARKER_FILES:
            marker_path = os.path.join(check_path, MARKER_FILE)
            if os.path.isfile(marker_path):
                return check_path

        check_path = os.path.dirname(check_path)



"""

METTA Cli package

"""

import logging

from mirantis.testing.metta.plugin import Type, Factory
from mirantis.testing.metta.environment import Environment

from .plugin import PluginCliPlugin
from .config import ConfigCliPlugin
from .environment import EnvironmentCliPlugin
from .fixtures import FixturesCliPlugin
from .output import OutputCliPlugin
from .provisioner import ProvisionerCliPlugin

logger = logging.getLogger('metta.cli')

METTA_PLUGIN_ID_CLI_PLUGIN = 'metta'
""" cli plugin_id for the plugin plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_PLUGIN)
def metta_plugin_factory_cli_plugin(
        environment: Environment, instance_id: str = ''):
    """ create an info cli plugin """
    return PluginCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_CONFIG = 'config'
""" cli plugin_id for the config plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_CONFIG)
def metta_plugin_factory_cli_config(
        environment: Environment, instance_id: str = ''):
    """ create a config cli plugin """
    return ConfigCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_ENVIRONMENT = 'environment'
""" cli plugin_id for the environment plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_ENVIRONMENT)
def metta_plugin_factory_cli_environment(
        environment: Environment, instance_id: str = ''):
    """ create an environment cli plugin """
    return EnvironmentCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_FIXTURES = 'fixtures'
""" cli plugin_id for the fixtures plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_FIXTURES)
def metta_plugin_factory_cli_fixtures(
        environment: Environment, instance_id: str = ''):
    """ create a fixtures cli plugin """
    return FixturesCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_OUTPUT = 'output'
""" cli plugin_id for the output plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_OUTPUT)
def metta_plugin_factory_output_config(
        environment: Environment, instance_id: str = ''):
    """ create a output cli plugin """
    return OutputCliPlugin(environment, instance_id)


METTA_PLUGIN_ID_CLI_PROVISIONER = 'provisioner'
""" cli plugin_id for the provisioner plugin """


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_CLI_PROVISIONER)
def metta_plugin_factory_provisioner_config(
        environment: Environment, instance_id: str = ''):
    """ create a provisioner cli plugin """
    return ProvisionerCliPlugin(environment, instance_id)


""" SetupTools EntryPoint BootStrapping """


def bootstrap(environment: Environment):
    """ METTA Bootstrapper - don't actually do anything """
    pass



import logging
from typing import Dict, Any

import json
import yaml

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase
from mirantis.testing.metta_cli.provisioner import ProvisionerGroup

logger = logging.getLogger('metta.cli.launchpad')


class LaunchpadCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands for aucnhpad provisioenrs if one hase been registered."""
        if self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_launchpad', exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'launchpad': LaunchpadGroup(self.environment)
                }
            }
        else:
            return {}


class LaunchpadGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """ Pick a matching provisioner """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_launchpad', instance_id=instance_id)
        else:
            # Get the highest priority provisioner
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_launchpad')

    def info(self, provisioner: str = '', deep: bool = False):
        """ get info about a provisioner plugin """
        fixture = self._select_provisioner(instance_id=provisioner)

        provisioner_info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                provisioner_info.update(fixture.plugin.info(True))

        return json.dumps(provisioner_info, indent=2)

    def hosts(self, provisioner: str = '', deep: bool = False):
        """ list the hosts in the cluster """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        config = client.describe_config()

        if deep:
            list = [host for host in config['spec']['hosts']]
        else:
            list = []
            for host in config['spec']['hosts']:
                list_host = {
                    'role': host['role']
                }
                if 'ssh' in host:
                    list_host.update({
                        'is_windows': False,
                        'address': host['ssh']['address']
                    })
                if 'winrm' in host:
                    list_host.update({
                        'is_windows': True,
                        'address': host['winrm']['address']
                    })

                list.append(list_host)

        return json.dumps(list, indent=2)

    def exec(self, cmd: str, provisioner: str = '',
             host: int = 0, interactive: bool = False):
        """ Exec a command """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        cmds = cmd.split(' ')

        client.exec(host_index=host, cmds=cmds)

    def exec_interactive(self, cmd: str, provisioner: str = '',
                         host: int = 0, interactive: bool = False):
        """ Exec a command """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        cmds = cmd.split(' ')

        client.exec_interactive(host_index=host, cmds=cmds)

    def connect(self, provisioner: str = '',
                host: int = 0, interactive: bool = False):
        """ Exec a command """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        client.exec_interactive(host_index=host, cmds=[])

    def fixtures(self, provisioner: str = ''):
        """ List all fixtures for this provisioner """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')
        list = [{
            'type': fixture.type.value,
            'plugin_id': fixture.plugin_id,
            'instance_id': fixture.instance_id,
            'priority': fixture.priority,
        } for fixture in provisioner.get_fixtures().to_list()]

        json.dumps(list, indent=2)

    def client_config(self, provisioner: str = ''):
        """ get the rendered config from the client """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin
        client = plugin.client

        return json.dumps(client.describe_config(), indent=2)

    def version(self, provisioner: str = ''):
        """ Output a launchpad cli report """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.client.version()

    def describe(self, report: str, provisioner: str = ''):
        """ Output a launchpad cli report """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.client.describe(report)

    def fixtures(self, provisioner: str = '', type: Type = None, plugin_id: str = '',
                 instance_id: str = ''):
        """ List all outputs """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')

        type = Type.from_string(type) if type else None

        list = [{
            'type': fixture.type.value,
            'plugin_id': fixture.plugin_id,
            'instance_id': fixture.instance_id,
            'priority': fixture.priority,
        } for fixture in provisioner.get_fixtures(type=type, plugin_id=plugin_id, instance_id=instance_id).to_list()]

        return json.dumps(list, indent=2)

    def config_file(self, provisioner: str = ''):
        """ Dump the config file """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin

        try:
            with open(provisioner.config_file) as f:
                config_contents = yaml.load(f)
        except FileNotFoundError as e:
            raise ValueError("No config file was found: {}".format(e)) from e

        return json.dumps(config_contents, indent=2)

    def write_config(self, provisioner: str = ''):
        """ write config to file """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner._write_launchpad_file()

    def client_bundle(self, provisioner: str = '',
                      user: str = 'admin', reload: bool = False):
        """ Tell Launchpad to download the client bundle """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin

        logger.info("Downloading client bungle for user: {}".format(user))
        try:
            bundle = plugin._mke_client_bundle(user=user, reload=reload)
        except Exception as e:
            raise Exception("Launchpad command failed : {}".format(e)) from e

        return json.dumps(bundle, indent=2)

    def prepare(self, provisioner: str = ''):
        """ Run provisioner prepare """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.prepare()

    def apply(self, provisioner: str = '', debug: (bool) = False):
        """ Run provisioner apply """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.apply(debug=debug)

    def destroy(self, provisioner: str = '', quick: bool = False):
        """ Run provisioner destroy """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.destroy(quick=quick)



"""

DUMMY Plugin testing.

As toolbox operations are quite context sensitive, and could involve creating
infrastructure, we don't have unit testing for them

Here we test the overall toolbox workflows using dummy plugins.

"""
import logging
import unittest

from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
# METTA components we will need to access the toolbox
from mirantis.testing.metta import new_environment, environment_names, get_environment
from mirantis.testing.metta.plugin import Type, Factory
# We import this so that we don't need to guess on plugin_ids, but not needed
from mirantis.testing.metta_dummy import METTA_PLUGIN_ID_DUMMY

# Imports used for type hinting
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
# imports used only for config creation for testing
from mirantis.testing.metta.provisioner import METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL, METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL
from mirantis.testing.metta.client import METTA_CLIENT_CONFIG_CLIENTS_LABEL, METTA_CLIENT_CONFIG_CLIENT_LABEL
from mirantis.testing.metta.workload import METTA_WORKLOAD_CONFIG_WORKLOADS_LABEL, METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL
# imports used only for type testing
from mirantis.testing.metta_dummy.provisioner import DummyProvisionerPlugin
from mirantis.testing.metta_dummy.client import DummyClientPlugin
from mirantis.testing.metta_dummy.workload import DummyWorkloadPlugin

logger = logging.getLogger("test_dummy")
logger.setLevel(logging.INFO)

""" Contents of test config files used as the source for a config object """

""" TESTS """


class PluginConstruction(unittest.TestCase):

    def _dummy_environment(self, name: str) -> Environment:
        """ Create an environment object, add the common config source data """
        if not name in environment_names():
            new_environment(
                name=name,
                additional_metta_bootstraps=['metta_dummy'])

        return get_environment(name=name)

    def test_1_construct_dicts(self):
        """ some sanity tests on constructors based on dicts """
        environment = self._dummy_environment('test_1')

        plugin_dict = {
            'plugin_id': METTA_PLUGIN_ID_DUMMY,
        }

        self.assertIsInstance(
            environment.add_fixture_from_dict(
                type=Type.PROVISIONER, plugin_dict=plugin_dict).plugin,
            DummyProvisionerPlugin)
        self.assertIsInstance(
            environment.add_fixture_from_dict(type=Type.CLIENT, plugin_dict=plugin_dict).plugin, DummyClientPlugin)
        self.assertIsInstance(
            environment.add_fixture_from_dict(
                type=Type.WORKLOAD, plugin_dict=plugin_dict).plugin,
            DummyWorkloadPlugin)

        plugins_dict = {
            'one': plugin_dict,
            'two': plugin_dict,
            'three': plugin_dict,
        }
        provisioners = environment.add_fixtures_from_dict(type=Type.PROVISIONER,
                                                          plugin_list=plugins_dict)

        self.assertIsInstance(provisioners, Fixtures)
        self.assertEqual(len(provisioners), 3)

        one = provisioners.get_plugin(instance_id='one')
        two = provisioners.get_plugin(instance_id='two')

        self.assertIsInstance(one, DummyProvisionerPlugin)
        self.assertIsInstance(two, DummyProvisionerPlugin)
        self.assertEqual(
            provisioners.get_plugin(
                type=Type.PROVISIONER).instance_id,
            'one')
        self.assertEqual(provisioners.get_plugin(type=Type.PROVISIONER), one)

    def test_2_construct_config(self):
        """ test that we can construct plugins from config """
        environment = self._dummy_environment('test_2')

        plugin_dict = {
            'plugin_id': METTA_PLUGIN_ID_DUMMY,
        }
        plugins_dict = {
            'one': plugin_dict,
            'two': plugin_dict,
            'three': plugin_dict,
        }

        environment.config.add_source(PLUGIN_ID_SOURCE_DICT, priority=80).set_data({
            METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL: plugins_dict,
            METTA_CLIENT_CONFIG_CLIENTS_LABEL: plugins_dict,
            METTA_WORKLOAD_CONFIG_WORKLOADS_LABEL: plugins_dict,

            METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL: plugin_dict,
            METTA_CLIENT_CONFIG_CLIENT_LABEL: plugin_dict,
            METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL: plugin_dict
        })

        self.assertIsInstance(
            environment.add_fixture_from_config(
                type=Type.PROVISIONER,
                label=METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL).plugin,
            DummyProvisionerPlugin)
        self.assertIsInstance(
            environment.add_fixture_from_config(
                type=Type.CLIENT, label=METTA_CLIENT_CONFIG_CLIENT_LABEL).plugin,
            DummyClientPlugin)
        self.assertIsInstance(
            environment.add_fixture_from_config(
                type=Type.WORKLOAD, label=METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL).plugin,
            DummyWorkloadPlugin)

        provisioners = environment.add_fixtures_from_config(
            type=Type.PROVISIONER, label=METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL)

        self.assertIsInstance(provisioners, Fixtures)
        self.assertEqual(len(provisioners), 3)

        two = provisioners.get_plugin(instance_id='two')

        self.assertIsInstance(two, DummyProvisionerPlugin)
        self.assertEqual(
            provisioners.get_plugin(
                type=Type.PROVISIONER).instance_id,
            'one')

    def test_3_mixedfixtures_flat(self):
        """ test building mixed fixtures from one config """
        environment = self._dummy_environment('test_3')
        environment.config.add_source(PLUGIN_ID_SOURCE_DICT, instance_id='fixtures', priority=80).set_data({
            'fixtures': {
                'prov': {
                    'type': Type.PROVISIONER.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY
                },

                'cl1': {
                    'type': Type.CLIENT.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY
                },
                'cl2': {
                    'type': Type.CLIENT.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY
                },

                'work1': {
                    'type': Type.WORKLOAD.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY,
                    'priority': environment.plugin_priority() - 10
                },
                'work2': {
                    'type': Type.WORKLOAD.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY,
                    'priority': environment.plugin_priority() + 10
                },
                'work3': {
                    'type': Type.WORKLOAD.value,
                    'plugin_id': METTA_PLUGIN_ID_DUMMY
                },
            },
        })

        environment.add_fixtures_from_config(label='fixtures')

        cl2 = environment.fixtures.get_plugin(
            type=Type.CLIENT, instance_id='cl2')
        self.assertEqual(cl2.instance_id, 'cl2')

        wls = environment.fixtures.get_plugins(type=Type.WORKLOAD)

        self.assertEqual(len(wls), 3)
        self.assertEqual(wls[0].instance_id, 'work2')
        self.assertEqual(wls[2].instance_id, 'work1')



import logging
from typing import Any, List, Dict

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.plugin import Factory as Factory, Type as Type

from .launchpad import LaunchpadClient
from .provisioner import LaunchpadProvisionerPlugin, METTA_LAUNCHPAD_CONFIG_LABEL, METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA
from .exec_client import LaunchpadExecClientPlugin, METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID as EXEC_CLIENT_PLUGIN_ID
from .cli import LaunchpadCliPlugin

""" GENERATING CONFIG  """

METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID = "metta_launchpad"
METTA_LAUNCHPAD_CLI_PLUGIN_ID = "metta_launchpad"
METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID = EXEC_CLIENT_PLUGIN_ID

""" provisioner plugin_id for the plugin """


@Factory(type=Type.PROVISIONER,
         plugin_id=METTA_LAUNCHPAD_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_launchpad(
        environment: Environment, instance_id: str = "", label: str = METTA_LAUNCHPAD_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
    """ create a launchpad provisioner plugin """
    return LaunchpadProvisionerPlugin(environment, instance_id, label, base)


@Factory(type=Type.CLIENT, plugin_id=METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID)
def metta_terraform_factory_cliexec_client_launchpad(
        environment: Environment, instance_id: str = '', client: LaunchpadClient = None):
    """ create an launchpad exec client plugin """
    return LaunchpadExecClientPlugin(environment, instance_id, client)


@Factory(type=Type.CLI, plugin_id=METTA_LAUNCHPAD_CLI_PLUGIN_ID)
def metta_terraform_factory_cli_launchpad(
        environment: Environment, instance_id: str = ''):
    """ create an launchpad cli plugin """
    return LaunchpadCliPlugin(environment, instance_id)


""" METTA BOOTSTRAPPERS """


def bootstrap(environment: Environment):
    """ metta configerus bootstrap

    Currently we only use this to import plugins.

    Parameters:
    -----------

    env (Environment) : an environment which should have validation config added
        to.

    """
    pass



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

from .output import OutputGroup

logger = logging.getLogger('metta.cli.fixtures')


class FixturesCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'fixture': FixturesGroup(self.environment)
        }


class FixturesGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def list(self, include_cli_plugins: bool = False):
        """ List all fixture instance_ids """
        list = [fixture.instance_id for fixture in self.environment.fixtures.get_fixtures(
        ).to_list() if include_cli_plugins or fixture.type is not Type.CLI]

        return json.dumps(list, indent=2)

    def info(self, type: str = '', plugin_id: str = '',
             instance_id: str = '', deep: bool = False, include_cli_plugins: bool = False, exception_if_missing: bool = False):
        """ Info for all fixtures """

        if type:
            type = Type.from_string(type)
        else:
            type = None

        list = []
        for fixture in self.environment.fixtures.get_fixtures(
                type=type, plugin_id=plugin_id, instance_id=instance_id).to_list():
            if not include_cli_plugins and fixture.type is Type.CLI:
                continue

            info = {
                'fixture': {
                    'type': fixture.type.value,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }

            if deep and hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            list.append(info)

        if exception_if_missing and len(list) == 0:
            raise ValueError("No matching fixtures")

        return json.dumps(list, indent=2)



"""

metta Docker

metta contrib package for docker functionality, specifically for registering
a Docker client plugin.

"""

from typing import Any

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .client import DockerPyClientPlugin
from .run_workload import DockerPyRunWorkloadPlugin, DOCKER_RUN_WORKLOAD_CONFIG_LABEL, DOCKER_RUN_WORKLOAD_CONFIG_BASE

METTA_PLUGIN_ID_DOCKER_CLIENT = 'metta_docker'
""" client plugin_id for the metta dummy plugin """


@Factory(type=Type.CLIENT, plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT)
def metta_plugin_factory_client_dockerpy(
    environment: Environment, instance_id: str = '', host: str = '', cert_path: str = '', tls_verify: bool = True,
        compose_tls_version: str = 'TLSv1_2', version: str = 'auto'):
    """ create an metta client for dockerpy """
    return DockerPyClientPlugin(environment, instance_id=instance_id,
                                host=host, cert_path=cert_path, tls_verify=tls_verify, compose_tls_version=compose_tls_version, version=version)


METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD = 'metta_docker_run'
""" workload plugin_id for the docker run plugin """


@Factory(type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD)
def metta_plugin_factory_workload_docker_run(
        environment: Environment, instance_id: str = '', label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL, base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE):
    """ create a docker run workload plugin """
    return DockerPyRunWorkloadPlugin(
        environment, instance_id, label=label, base=base)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_Docker bootstrap

    We dont't take any action.  Our purpose is to run the above factory
    decorators to register our plugins.

    """
    pass



import logging
from typing import Dict, Any

import json
import yaml

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

logger = logging.getLogger('metta.cli.kubernetes')


class KubernetesCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands for aucnhpad provisioenrs if one hase been registered."""
        if self.environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id='metta_kubernetes', exception_if_missing=False) is not None:

            return {
                'contrib': {
                    'kubernetes': KubernetesGroup(self.environment)
                }
            }

        else:
            return {}


class KubernetesGroup():

    def __init__(self, environment: Environment):
        self._environment = environment

        if self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_yaml', exception_if_missing=False) is not None:
            self.yaml = KubernetesYamlWorkloadGroup(self._environment)
        if self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_helm', exception_if_missing=False) is not None:
            self.helm = KubernetesHelmWorkloadGroup(self._environment)

    def _select_client(self, instance_id: str = ''):
        """ Pick a matching client """
        if instance_id:
            return self._environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id='metta_kubernetes', instance_id=instance_id)
        else:
            # Get the highest priority workload
            return self._environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id='metta_kubernetes')

    def info(self, workload: str = '', deep: bool = False):
        """ get info about a client plugin """
        fixture = self._select_client(instance_id=workload)

        collect_info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return json.dumps(collect_info, indent=2)

    def readyz(self, workload: str = '', verbose: bool = False):
        """ get kubernetes readiness info from the plugin """
        plugin = self._select_client(instance_id=workload).plugin

        try:
            return json.dumps(plugin.readyz(verbose=verbose),
                              indent=2, default=lambda x: "{}".format(x))

        except Exception as e:
            raise RuntimeError('Kubernetes is not ready') from e

    def livez(self, workload: str = '', verbose: bool = False):
        """ get kubernetes livez info from the plugin """
        plugin = self._select_client(instance_id=workload).plugin

        try:
            return json.dumps(plugin.livez(verbose=verbose),
                              indent=2, default=lambda x: "{}".format(x))

        except Exception as e:
            raise RuntimeError('Kubernetes is not ready') from e

    def connect_service_proxy(self, namespace: str,
                              service: str, workload: str = ''):
        """ create a service proxy """
        plugin = self._select_client(instance_id=workload).plugin

        try:
            CoreV1Api = plugin.get_api('CoreV1Api')
            sc = CoreV1Api.connect_post_namespaced_service_proxy(
                namespace=namespace, name=service)

            return json.dumps(sc, indent=2, default=lambda x: "{}".format(x))

        except Exception as e:
            raise RuntimeError(
                'Exception trying to open the service proxy') from e


class KubernetesYamlWorkloadGroup():

    def __init__(self, environment: Environment):
        self._environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """ Pick a matching workload fixture """
        if instance_id:
            return self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_yaml', instance_id=instance_id)
        else:
            # Get the highest priority workload
            return self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_yaml')

    def info(self, workload: str = '', deep: bool = False):
        """ get info about a yaml workload plugin """
        fixture = self._select_fixture(instance_id=workload)

        collect_info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return json.dumps(collect_info, indent=2)

    def apply(self, workload: str = ''):
        """ Run workload apply """
        workload = self._select_fixture(instance_id=workload).plugin
        instance = workload.create_instance(self._environment.fixtures)

        objects = instance.apply()

        return json.dumps(objects, indent=2, default=lambda x: "{}".format(x))

    def destroy(self, workload: str = ''):
        """ Run workload destroy """
        workload = self._select_fixture(instance_id=workload).plugin
        instance = workload.create_instance(self._environment.fixtures)

        destroy = instance.destroy()

        return json.dumps(destroy, indent=2)


class KubernetesHelmWorkloadGroup():

    def __init__(self, environment: Environment):
        self._environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """ Pick a matching workload fixture """
        if instance_id:
            return self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_helm', instance_id=instance_id)
        else:
            # Get the highest priority workload
            return self._environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id='metta_kubernetes_helm')

    def info(self, workload: str = '', deep: bool = False):
        """ get info about a helm workload plugin """
        fixture = self._select_fixture(instance_id=workload)

        collect_info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                collect_info.update(fixture.plugin.info(True))

        return json.dumps(collect_info, indent=2)

    def apply(self, workload: str = '', wait: bool = True,
              debug: bool = False):
        """ Run helm workload apply """
        workload = self._select_fixture(instance_id=workload).plugin
        instance = workload.create_instance(self._environment.fixtures)

        objects = instance.apply(wait=wait, debug=debug)

        return json.dumps(objects, indent=2, default=lambda x: "{}".format(x))

    def destroy(self, workload: str = '', debug: bool = False):
        """ Run helm workload destroy """
        workload = self._select_fixture(instance_id=workload).plugin
        instance = workload.create_instance(self._environment.fixtures)

        destroy = instance.destroy(debug=debug)

        return json.dumps(destroy, indent=2)



import logging
import os
import sys

from mirantis.testing.metta import environment_names, get_environment, discover, new_environment
from mirantis.testing.metta.plugin import Type, Factory


logger = logging.getLogger('metta.cli.base')

UCC_CLI_FIXTURE_KEY_CONFIG = 'config'

FILES = {
    'metta': 'metta.py',
    'mettac': 'mettac.py',
}


class Base:
    """ The Metta CLI program

    This CLI let's you interact with a metta environment for the purpose of
    introspection, debugging and manual interaction.

    All groups and commands come from metta CLI plugins.

    """

    def __init__(self, environment: str = '', state: str = ''):
        """

        Parameters:
        -----------

        environment (str) : Environment name in case you want to switch to an
            alternate environment.

        state (str) : Environment state to consider active.  Will throw an error
            if the state doesn't exist for the selected environment.

        """

        # Try to make an environment using
        discover()

        try:
            if environment == '':
                environment = environment_names()[0]
        except (KeyError, IndexError) as e:
            logger.warn(
                "No environment object has been defined (making one now.) Are you in a project folder?")
            environment = 'empty'
            new_environment(environment)

        try:
            self._environment = get_environment(environment)

            if state:
                self._environment.set_state(state)
        except KeyError:
            raise ValueError(
                "Could not load environment '{}', not found; Existing environments: {}".format(
                    environment, environment_names()))

        # collect any comands from all discovered cli plugins
        self._collect_commands()

    def _collect_commands(self):
        """ collect commands from all cli plugins

        Create an instance of any registered cli plugin.
        From the plugin, collect the commands and add each command to this
        object directly, so that Fire can see them.

        """

        plugin_list = {}
        for plugin_id in Factory.registry[Type.CLI.value]:
            plugin_list[plugin_id] = {
                'type': Type.CLI.value,
                'plugin_id': plugin_id
            }

        for plugin in self._environment.add_fixtures_from_dict(
                plugin_list=plugin_list, type=Type.CLI).get_plugins():
            logger.info("loading cli plugin: {}".format(plugin_id))

            if hasattr(plugin, 'fire'):
                try:
                    commands = plugin.fire()
                except TypeError as e:
                    raise NotImplementedError(
                        "Plugin {} did not implement the correct fire(fixtures) interface: {}".format(
                            plugin_id, e)) from e

                if not isinstance(commands, dict):
                    raise ValueError(
                        "Plugin returned invalid commands : {}".format(commands))

                ValueError(
                    "Plugin returned invalid commands : {}".format(commands))

                for (command_name, command) in commands.items():
                    logger.debug(
                        "adding cli plugin command: {}->{}".format(plugin_id, command_name))

                    # if the command name already exists and is a dict then
                    # maybe we should merge them
                    if hasattr(self, command_name) and isinstance(
                            getattr(self, command_name), dict) and isinstance(command, dict):
                        getattr(self, command_name).update(command)
                        continue

                    setattr(self, command_name, command)



import logging
import pkg_resources

import appdirs

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH as CONFIGERUS_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT as CONFIGERUS_SOURCE_DICT

from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta_mirantis.common")

METTA_MIRANTIS_COMMON_NAME = "metta-mirantis"
""" used for some path building for config sources """
METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS = 35
""" Config source priority for added common config """
METTA_MIRANTIS_PATH = pkg_resources.resource_filename(
    "mirantis.testing.metta_mirantis", "")
""" Path to the metta mirantis project root """
METTA_MIRANTIS_CONFIG_PATH = pkg_resources.resource_filename(
    'mirantis.testing.metta_mirantis', "config")
""" Path to the metta mirantis config preset configurations """


def add_common_config(environment: Environment):
    """ Add some common configuration sources

    The following could be added:

    1. a path config key for a path to this module.  This is used to make it
       easier to build paths to things like the terraform plans in this package
    2. add this module's ./config path as a config source.

    Parameters:
    -----------

    environment (Environment) : METTA environment to be modified

    """

    # If we haven't already done it, add the metta_mirantis/config path as a
    # config source.  This allows us to provide config defaults in this module
    metta_mirantis_config_instance_id = "{}-config".format(
        METTA_MIRANTIS_COMMON_NAME)
    if not environment.config.has_source(metta_mirantis_config_instance_id):
        environment.config.add_source(
            CONFIGERUS_SOURCE_PATH,
            metta_mirantis_config_instance_id,
            priority=METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS).set_path(METTA_MIRANTIS_CONFIG_PATH)

    # Add a "{path:metta_mirantis}" config value so that paths to this module can
    # be built.  We use this to allow file resources from this module to be used
    # as a part of config.  For example we build paths to the terraform plans in
    # this module.
    metta_mirantis_paths_instance_id = "{}-paths".format(
        METTA_MIRANTIS_COMMON_NAME)
    if not environment.config.has_source(metta_mirantis_paths_instance_id):
        environment.config.add_source(CONFIGERUS_SOURCE_DICT, metta_mirantis_paths_instance_id, priority=METTA_MIRANTIS_COMMON_DEFAULT_SOURCE_PRIORITY_DEFAULTS).set_data({
            environment.config.paths_label(): {
                METTA_MIRANTIS_COMMON_NAME: METTA_MIRANTIS_PATH
            }
        })




import kubernetes
import logging
import re
import time
from typing import Dict

from mirantis.testing.metta.client import ClientBase

logger = logging.getLogger('metta.contrib.kubernetes.client.kubeapi')


class KubernetesApiClientPlugin(ClientBase):
    """ metta Client plugin for Kubernetes

    Construction:
    -------------

    As an metta Plugin, the class constructor is used to initialize the instance
    as a plugin.  This means that we lose the constructor, but we gain access
    to an metta.config.Config object.

    We use .args() to configure the instance and create the client, as per the
    client plugin standard.

    To use:
    -------

    2. Ask metta.client.make_client for an instance of the plugin
        a. it is more expected that a provisioner will provide the client directly
        b. you can ask for the plugin directly from metta.plugin.get_plugin()
        c. you can use metta.new_clients_from_config() if you have some configuration
           with the needed args, and it will build clients for you
    3. Configure using .args() adding a kubeconfig file (with activated context)
       if that wasn't done for you when you received the client
    4. Ask the client for a kubernetes API version client
       (such as get_api('CoreV1Api') => CoreV1Api)
    5. Use the kubernetes API client as normal

    ```
    import mirantis.testing.metta as metta
    import mirantis.testing.metta.client as metta_client
    import mirantis.testing.metta_kubernetes as metta_kubernetes

    config = metta.new_config()
    client = metta_client.make_client(metta_kubernetes.METTA_PLUGIN_ID_KUBERNETES_CLIENT, config, 'my-k8-instance')

    client.args('path/to/k8file')

    coreV1 = client.get_api('CoreV1Api')
    ns = coreV1.read_namespace(name="kube-system")
    print("NS: {}".format(ns))
    ```

    Why use this:
    -------------

    It is cumbersome to use this if you already have access to the kubeconfig, but
    it fits into the plugin system, and therefore auto-loading and auto-configuring
    is possible with this implementation.
    Effectively, it is not easier to use this plu-gin over the native K8 client unless
    you consider that a provisioner can provide you this plugin already configured
    with its own information.

    This plugin is particularly usefull as it can be used for the helm and deployment
    workload plugins directly, which know how to use it to apply workloads to a
    kubernetes cluster.

    """

    def __init__(self, environment, instance_id, kube_config_file: str = ''):
        """ Run the super constructor but also set class properties

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------

        config_file (str): String path to the kubernetes config file to use

        """
        super(ClientBase, self).__init__(environment, instance_id)

        logger.debug("Creating Kuberentes client from config file")
        self.api_client = kubernetes.config.new_client_from_config(
            config_file=kube_config_file)
        """ Kubernetes api client """

        self.config_file = kube_config_file
        """ Kube config file, in case you need to steal it. """

    def info(self):
        """ Return dict data about this plugin for introspection """
        return {
            'kubernetes': {
                'config_file': self.config_file
            }
        }

    def watch(self):
        """ Get a kubernetes watch instance """
        return kubernetes.watch.Watch()

    def get_api(self, api: str):
        """ Get an kubernetes API """
        if hasattr(kubernetes.client, api):
            return getattr(kubernetes.client, api)(self.api_client)

        raise KeyError("Unknown API requested: {}".format(api))

    def utils_create_from_yaml(self, yaml_file: str, **kwargs):
        """ Run a kube apply from a yaml file """
        return kubernetes.utils.create_from_yaml(
            k8s_client=self.api_client, yaml_file=yaml_file, **kwargs)

    def utils_create_from_dict(self, data: Dict, **kwargs):
        """ Run a kube apply from dict of K8S yaml """
        return kubernetes.utils.create_from_dict(
            k8s_client=self.api_client, data=data, **kwargs)

    def ready_wait(self, timeout: int = 30, period: int = 1):
        """ Wait until kubernetes is ready before returning """
        err = None
        while timeout > 0:
            try:
                ready = self.readyz()
                return
            except Exception as e:
                err = e
                time.sleep(period)
                timeout = timeout - period
                continue

        raise RuntimeError('Timed out waiting for kubernetes to become ready') from err

    def readyz(self, verbose: bool = False):
        """ check the general readyz endpoint

        Returns:
        --------

        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------

        Will raise an exception is kubernetes isn't avaialble

        """
        return self._interpret_z_response('/readyz')

    def livez(self, verbose: bool = False):
        """ check the general livez endpoint

        Returns:
        --------

        Dict of service:status_dict with values symbol (+) and string ok value

        Raises:
        -------

        Will raise an exception is kubernetes isn't avaialble

        """
        return self._interpret_z_response('/livez')

    def _interpret_z_response(self, endpoint: str, method: str = 'GET', params: Dict[str, str] = {}) -> Dict[str, Dict[str, str]]:
        """ interpret that readyz/livez response format into a dict """
        INTERPRET_REGEX = re.compile(
            r'^\[(?P<symbol>[+-])\](?P<name>\S+)\s{1}(?P<ok>\w+)$')

        # this will produce an exception if K8s is not ready
        response = self.api_client.call_api(
            method=method,
            resource_path=endpoint,
            query_params=params,
            _preload_content=False)[0]

        interpreted = {}
        for line in response.read().decode("utf-8") .split('\n'):
            match = INTERPRET_REGEX.fullmatch(line)
            if match:
                interpreted[match.group('name')] = {
                    'symbol': match.group('symbol'),
                    'ok': match.group('ok')
                }

        return interpreted



from typing import Any

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .provisioner import TestkitProvisionerPlugin, METTA_PLUGIN_ID_TESTKIT_PROVISIONER, TESTKIT_PROVISIONER_CONFIG_LABEL, TESTKIT_PROVISIONER_CONFIG_BASE
from .cli import TestkitCliPlugin, METTA_PLUGIN_ID_TESTKIT_CLI


@Factory(type=Type.PROVISIONER,
         plugin_id=METTA_PLUGIN_ID_TESTKIT_PROVISIONER)
def metta_plugin_factory_provisioner_testkit(environment: Environment, instance_id: str = '', label: str = TESTKIT_PROVISIONER_CONFIG_LABEL, base: Any = TESTKIT_PROVISIONER_CONFIG_BASE):
    """ create an metta litmuschaos workload plugin """
    return TestkitProvisionerPlugin(
        environment, instance_id, label=label, base=base)


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_TESTKIT_CLI)
def metta_plugin_factory_cli_testkit(environment: Environment, instance_id: str = ''):
    """ create an litmuschaos cli plugin """
    return TestkitCliPlugin(environment, instance_id)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_testkit bootstrap

    Currently we only use this to import plugins.

    Parameters:
    -----------

    env (Environment) : an environment which should have validation config added
        to.

    """
    pass



"""

Metta CLI : User plugin

Examine and declare user configuration for Metta

"""
import logging
import appdirs
import os
from typing import Dict, Any, List

import json
import yaml

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

from .common_config import METTA_COMMON_APP_NAME

logger = logging.getLogger('metta.cli.user')


class UserCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'user': UserGroup(self.environment)
        }


class UserGroup():

    def __init__(self, environment: Environment):
        self._environment = environment

    def info(self):
        """ Output any user related information """
        user_config = self._environment.config.load('user')
        user_config_dir = self._user_path()

        info = {
            'config': user_config.get(default={}),
            'system': {
                'path': user_config_dir,
                'exists': os.path.isdir(user_config_dir)
            }
        }

        return json.dumps(info, indent=2)

    def init(self):
        """ Make a User config source that will be global for this system """
        user_config_dir = self._user_path()
        if self._user_path_exists():
            raise RuntimeError(
                "User has already been initialized on this system: {}".format(user_config_dir))

        os.mkdir(user_config_dir)

    # @TODO generate a more abstract method for setting values
    def set_id(self, id: str):
        """ set the user id for the user on the system """
        if not self._user_path_exists():
            raise RuntimeError(
                "No user configuration has been initialized on this system.  Run `init` first")

        user_config_dir = self._user_path()

        config_path_user = os.path.join(user_config_dir, 'user.yml')
        try:
            with open(config_path_user, 'r') as user_config_file:
                user_config = yaml.load(user_config_file)
        except FileNotFoundError:
            user_config = {}

        user_config['id'] = id

        with open(config_path_user, 'w') as user_config_file:
            yaml.dump(user_config, user_config_file)

    def _user_path(self):
        """ path to user specific configuration """
        # a user config path (like ~/.config/metta) may contain config
        return appdirs.user_config_dir(METTA_COMMON_APP_NAME)

    def _user_path_exists(self):
        """ boolean check on whether or not the user config path exists on the system """
        # a user config path (like ~/.config/metta) may contain config
        user_config_dir = self._user_path()
        return os.path.isdir(user_config_dir)



"""

Ansible METTA provisioner plugin

"""

import logging
import json
import os
import subprocess
from typing import Dict, List, Any

from configerus.loaded import LOADED_KEY_ROOT
from mirantis.testing.metta.plugin import METTAPlugin, Type
from mirantis.testing.metta.fixtures import Fixtures, UCCTFixturesPlugin, METTA_FIXTURES_CONFIG_FIXTURES_LABEL
from mirantis.testing.metta.provisioner import ProvisionerBase
from mirantis.testing.metta.output import OutputBase
from mirantis.testing.metta_common import METTA_PLUGIN_ID_OUTPUT_DICT, METTA_PLUGIN_ID_OUTPUT_TEXT

logger = logging.getLogger('metta.contrib.provisioner:ansible')

ANSIBLE_PROVISIONER_CONFIG_LABEL = 'ansible'
""" config label loading the ansible config """
ANSIBLE_PROVISIONER_CONFIG_PLAN_PATH_KEY = 'plan.path'
""" config key for the ansible plan path """
ANSIBLE_PROVISIONER_CONFIG_STATE_PATH_KEY = 'state.path'
""" config key for the ansible state path """
ANSIBLE_PROVISIONER_CONFIG_VARS_KEY = 'vars'
""" config key for the ansible vars Dict, which will be written to a file """
ANSIBLE_PROVISIONER_CONFIG_VARS_PATH_KEY = 'vars_path'
""" config key for the ansible vars file path, where the plugin will write to """
ANSIBLE_PROVISIONER_DEFAULT_VARS_FILE = 'metta_ansible.tfvars.json'
""" Default vars file if none was specified """
ANSIBLE_PROVISIONER_DEFAULT_STATE_SUBPATH = 'metta-state'
""" Default vars file if none was specified """


class AnsibleProvisionerPlugin(ProvisionerBase, UCCTFixturesPlugin):
    """ Ansible provisioner plugin

    Provisioner plugin that allows control of and interaction with a ansible
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a ansible binary, so you have to install
       ansible in the environment

    ## Usage

    ### Plan

    The plan must exists somewhere on disk, and be accessible.

    You must specify the path and related configuration in config, which are read
    in the .prepare() execution.

    ### Vars/State

    This plugin reads TF vars from config and writes them to a vars file.  We
    could run without relying on vars file, but having a vars file allows cli
    interaction with the cluster if this plugin messes up.

    You can override where Ansible vars/state files are written to allow sharing
    of a plan across test suites.

    """

    def __init__(self, environment, instance_id,
                 label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties

        Interpret provided config and configure the object with all of the needed
        pieces for executing ansible commands

        """
        super(ProvisionerBase, self).__init__(environment, instance_id)

        logger.info("Preparing Ansible setting")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

        fixtures = self.environment.add_fixtures_from_config(
            label=self.config_label,
            base=[self.config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL])
        """ All fixtures added to this provisioner plugin. """
        UCCTFixturesPlugin.__init__(self, fixtures)

        self.ansible_config = self.environment.config.load(
            self.config_label)
        """ get a configerus LoadedConfig for the ansible label """

    def info(self):
        """ get info about a provisioner plugin """
        return {
            'plugin': {
                'config_label': self.config_label,
                'config_base': self.config_base
            }
        }

    def prepare(self, label: str = '', base: str = ''):
        """ Prepare the provisioner to apply resources

        Initial Provisioner plugin is expected to be of very low cost until
        prepare() is executed.  At this point the plugin should load any config
        and perform any validations needed.
        The plugin should not create any resources but it is understood that
        there may be a cost of preparation.

        Provisioners are expected to load a lot of config to self-program.
        Because of this, they allow passing of a configerus label for .load()
        and a base for .get() in case there is an alterante config source
        desired.

        """
        pass

    def apply(self):
        """ bring a cluster to the configured state """
        pass

    def destroy(self):
        """ remove all resources created for the cluster """
        pass



"""

metta Kubernetes

metta contrib functionality for Kubernetes.  In particular to provide plugins for
kubernetes clients and kubernetes workloads.

"""

import os
from typing import List, Any

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .kubeapi_client import KubernetesApiClientPlugin
from .deployment_workload import KubernetesDeploymentWorkloadPlugin, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE
from .yaml_workload import KubernetesYamlWorkloadPlugin, KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL, KUBERNETES_YAML_WORKLOAD_CONFIG_BASE
from .helm_workload import KubernetesHelmWorkloadPlugin, KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL, KUBERNETES_HELM_WORKLOAD_CONFIG_BASE
from .cli import KubernetesCliPlugin

METTA_PLUGIN_ID_KUBERNETES_CLIENT = 'metta_kubernetes'
""" client plugin_id for the metta kubernetes client plugin """


@Factory(type=Type.CLIENT, plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
def metta_plugin_factory_client_kubernetes(
        environment: Environment, instance_id: str = '', kube_config_file: str = ''):
    """ create an metta kubernetes client plugin """
    return KubernetesApiClientPlugin(
        environment, instance_id, kube_config_file)


METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD = 'metta_kubernetes_deployment'
""" workload plugin_id for the metta_kubernetes deployment plugin """


@Factory(type=Type.WORKLOAD,
         plugin_id=METTA_PLUGIN_ID_KUBERNETES_DEPLOYMENT_WORKLOAD)
def metta_plugin_factory_workload_kubernetes_deployment(
        environment: Environment, instance_id: str = '', label: str = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE):
    """ create an metta kubernetes spec workload plugin """
    return KubernetesDeploymentWorkloadPlugin(
        environment, instance_id, label=label, base=base)


METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD = 'metta_kubernetes_yaml'
""" workload plugin_id for the metta_kubernetes yaml plugin """


@Factory(type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_KUBERNETES_YAML_WORKLOAD)
def metta_plugin_factory_workload_kubernetes_yaml(environment: Environment, instance_id: str = '',
                                                  label: str = KUBERNETES_YAML_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_YAML_WORKLOAD_CONFIG_BASE):
    """ create an metta kubernetes spec workload plugin """
    return KubernetesYamlWorkloadPlugin(
        environment, instance_id, label=label, base=base)


METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD = 'metta_kubernetes_helm'
""" workload plugin_id for the metta_kubernetes helm plugin """


@Factory(type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD)
def metta_plugin_factory_workload_kubernetes_helm(environment: Environment, instance_id: str = '',
                                                  label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE):
    """ create an metta kubernetes spec workload plugin """
    return KubernetesHelmWorkloadPlugin(
        environment, instance_id, label=label, base=base)


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
def metta_terraform_factory_cli_kubernetes(
        environment: Environment, instance_id: str = ''):
    """ create a kubernetes cli plugin """
    return KubernetesCliPlugin(environment, instance_id)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_Kubernetes bootstrap

    We dont't take any action.  Our purpose is to run the above factory
    decorators to register our plugins.

    """
    pass



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

logger = logging.getLogger('metta.cli.testkit')

METTA_PLUGIN_ID_TESTKIT_CLI = 'testkit'

class TestkitCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        if self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_testkit', exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'testkit': TestkitGroup(self.environment)
                }
            }
        else:
            return {}


class TestkitGroup():

    def __init__(self, environment: Environment):
        self.environment = environment


    def _select_provisioner(self, instance_id: str = ''):
        """ Pick a matching provisioner (in case you have more than one testkit provisioner configured) """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_testkit', instance_id=instance_id)
        else:
            # Get the highest priority provisioner
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_testkit')


    def info(self, provisioner: str = '', deep: bool = False):
        """ get info about a provisioner plugin """
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                info.update(fixture.plugin.info(True))

        return json.dumps(info, indent=2)

    def prepare(self, provisioner: str = ''):
        """ Run the provisioner prepare : which for testkit currently does nothing """
        pass

    def apply(self, provisioner: str = ''):
        """ Run the provisioner apply, which runs testkit create """
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        return json.dumps(plugin.apply())

    def destroy(self, provisioner: str = ''):
        """ Run the provisioner destroy, which runs testkit system rm """
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        return json.dumps(plugin.destroy())



from typing import Any, List, Dict
import logging
import os
import subprocess
import yaml
from enum import Enum
from datetime import datetime

import kubernetes

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.workload import WorkloadBase

logger = logging.getLogger('metta.contrib.kubernetes.workload.helm')


class HelmWorkloadPlugin(WorkloadBase):
    """ Helm client for interacting with a kube client """


KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL = 'kubernetes'
""" default config label used to load the workload """
KUBERNETES_HELM_WORKLOAD_CONFIG_BASE = 'workload.helm'
""" default config key for creating a workload object """

KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VERSION = 'version'
""" config key for helm version """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS = 'repos'
""" config key for helm repos dict which need to be added """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART = 'chart'
""" config key for helm chart, either local path or http(s) """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET = 'set'
""" config key for helm chart values """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES = 'values'
""" config key for helm chart values that should be put into a values file """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH = 'file'
""" config key for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH = 'values.yml'
""" config default value for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE = 'namespace'
""" config key for namespace to install to """

KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE = 'default'
""" default namespace to install to if no namespace was passed """


class KubernetesHelmWorkloadPlugin(WorkloadBase):
    """ Kubernetes workload class """

    def __init__(self, environment, instance_id, label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL,
                 base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE):
        """ Run the super constructor but also set class properties

        This implements the args part of the client interface.

        Here we expect to receiv config pointers so that we can determine what
        helm workload to apply, and what values will be needed.

        Parameters:
        -----------

        config_file (str): String path to the kubernetes config file to use

        """
        WorkloadBase.__init__(self, environment, instance_id)

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """ Create a workload instance from a set of fixtures

        Parameters:
        -----------

        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """

        try:
            client = fixtures.get_plugin(
                type=Type.CLIENT, plugin_id='metta_kubernetes')
        except KeyError as e:
            raise NotImplementedError(
                "Workload could not find the needed client: {}".format('metta_kubernetes'))

        workload_config = self.environment.config.load(self.config_label)

        kubeconfig = client.config_file

        namespace = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE], default=KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE)

        chart = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART])

        set = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET], default={})
        values = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES], default={})
        file = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH], default=KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH)

        repos = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS], default={})

        return KubernetesHelmV3WorkloadInstance(
            kubeconfig=kubeconfig, namespace=namespace, name=self.instance_id, repos=repos, chart=chart, set=set, values=values, file=file)

    def info(self, deep: bool = False):
        """ Return dict data about this plugin for introspection """
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'repos': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS]),
                    'chart': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART]),
                    'values': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUES])
                },
                'required_fixtures': {
                    'kubernetes': {
                        'type': Type.CLIENT.value,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


class KubernetesHelmV3WorkloadInstance:

    def __init__(self, kubeconfig, namespace, name, repos, chart, set: Dict[str, str] = {},
                 values: Dict[str, Any] = {}, file: str = KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH):
        """

        Parameters:
        -----------

        kubeconfig (str) : Path to the kubeinfo file

        repos (dict[str, str]) : name->path for repos to add

        chart (str) : local or http(s) path to a chart

        set (dict[str, str]) : simple helm chart values that can be set using
           the --set options on the command line

        values (dict[str, Any]) : complex helm chart value data that needs to
            be put into a file
        file (str) : path to where we should put values yaml file

        """

        self.kubeconfig = kubeconfig
        self.name = name
        self.namespace = namespace
        self.repos = repos
        self.chart = chart

        self.set = set

        self.values = values
        self.file = file

        self.bin = 'helm'
        self.working_dir = '.'

    def apply(self, wait: bool = True, debug: bool = False):
        """ Apply the helm chart

        To make the helm apply method reusable, we always run an upgrade with
        the --install flag, which w orks for both the first install, and any
        upgrades run after install

        Parameters:
        -----------

        wait (bool) : ask the helm client to wait until resources are created
            before returning.

        debug (bool) : ask the helm client for verbose output

        """

        for repo_name, repo_url in self.repos.items():
            self._run(cmd=['repo', 'add', repo_name, repo_url])

        cmd = ['upgrade']
        cmd += [self.name, self.chart]

        cmd += ['--install', '--create-namespace']

        if wait:
            cmd += ['--wait']
        if debug:
            cmd += ['--debug']

        if len(self.set):
            # turn the set dict into '--set a=A,b=B,c=C'
            cmd += ['--set', ','.join(['{}={}'.format(name, value)
                                       for (name, value) in self.values.items()])]

        if len(self.values):
            # turn the values into a file, and add it to the command
            with open(self.file, 'w') as f:
                yaml.dump(self.values, f)

            cmd += ['--values', self.file]

        try:
            self._run(cmd=cmd)
        except Exception as e:
            raise RuntimeError(
                'Helm failed to install the relese: {}'.format(e)) from e

    def list(self, all: bool = False, failed: bool = False,
             deployed: bool = False, pending: bool = False):
        """ List all releases - Not instances specific but still useful """
        cmd = ['list', '--output={}'.format('yaml')]

        if all:
            cmd += ['--all']
        elif deployed:
            cmd += ['--deployed']
        elif failed:
            cmd += ['--failed']
        elif pending:
            cmd += ['--pending']

        list_str = self._run(cmd=cmd, return_output=True)
        if list_str:
            return yaml.safe_load(list_str)

        return []

    def destroy(self, debug: bool = False):
        """ remove an installed helm release

        Parameters:
        -----------

        debug (bool) : ask the helm client for verbose output

        """

        cmd = ['uninstall', self.name]

        if debug:
            cmd += ['--debug']

        self._run(cmd=cmd)

    def test(self):
        """ test an installed helm release

        This runs the helm client test command.

        """
        self._run(cmd=['test', self.name])

    def status(self):
        """ status of the installed helm release """
        return HelmReleaseStatus(yaml.safe_load(
            self._run(cmd=['status', self.name, '--output', 'yaml'], return_output=True)))

    def _run(self, cmd: List[str], return_output: bool = False):
        """ run a helm v3 command """

        cmd = [self.bin,
               '--kubeconfig={}'.format(self.kubeconfig),
               '--namespace={}'.format(self.namespace)] + cmd

        if return_output:
            logger.debug(
                "running launchpad command with output capture: %s",
                " ".join(cmd))
            exec = subprocess.run(
                cmd,
                cwd=self.working_dir,
                shell=False,
                stdout=subprocess.PIPE)
            exec.check_returncode()
            return exec.stdout.decode('utf-8')
        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            exec = subprocess.run(
                cmd,
                cwd=self.working_dir,
                check=True,
                text=True)
            exec.check_returncode()


class Status(Enum):
    """ A Helm Status enum """

    DEPLOYED = 'deployed'
    """ Helm Release has been deployed """


class HelmReleaseStatus:
    """ interpreted helm release status

    Used to formalize the response object from a status request

    """

    def __init__(self, status_reponse: dict):
        """ interpret status from a status response """
        self.name = status_reponse['name']
        self.version = status_reponse['version']
        self.namespace = status_reponse['namespace']

        # self.first_deployed = datetime.strptime(status_reponse['info']['first_deployed'], '%Y-%m-%dT%H:%M:%S.%fZ')
        # self.last_deployed = datetime.strptime(status_reponse['info']['last_deployed'], '%Y-%m-%dT%H:%M:%S.%fZ')
        self.deleted = status_reponse['info']['deleted']
        self.description = status_reponse['info']['description']
        self.status = Status(status_reponse['info']['status'])
        self.notes = status_reponse['info']['notes']

        self.config = status_reponse['config']

        self.manifect = status_reponse['manifest']



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase
from mirantis.testing.metta_cli.provisioner import ProvisionerGroup

from .provisioner import METTA_TERRAFORM_PROVISIONER_PLUGIN_ID

logger = logging.getLogger('metta.cli.terraform')

METTA_TERRAFORM_CLI_PLUGIN_ID = 'metta_terraform'
""" cli plugin_id for the info plugin """

class TerraformCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands

        Don't return any commands if there is no provisioner plugin available

        """
        if self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID, exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'terraform': TerraformGroup(self.environment)
                }
            }
        else:
            return {}


class TerraformGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """ Pick a matching provisioner """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID, instance_id=instance_id)
        else:
            # Get the highest priority provisioner
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID)

    def info(self, provisioner: str = '', deep: bool = False):
        """ get info about a provisioner plugin """
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            if hasattr(fixture.plugin, 'get_fixtures'):
                fixtures = {}
                for sub_fixture in fixture.plugin.get_fixtures().to_list():
                    fixture_info = {
                        'fixture': {
                            'type': sub_fixture.type.value,
                            'plugin_id': sub_fixture.plugin_id,
                            'instance_id': sub_fixture.instance_id,
                            'priority': sub_fixture.priority,
                        }
                    }
                    if hasattr(sub_fixture.plugin, 'info'):
                        plugin_info = sub_fixture.plugin.info()
                        if isinstance(plugin_info, dict):
                            fixture_info.update(plugin_info)
                    fixtures[sub_fixture.instance_id] = fixture_info
                info['fixtures'] = fixtures

        return json.dumps(info, indent=2, default=lambda X: "{}".format(X))

    def fixtures(self, provisioner: str = ''):
        """ List all fixtures for this provisioner """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')
        list = [{
            'type': fixture.type.value,
            'plugin_id': fixture.plugin_id,
            'instance_id': fixture.instance_id,
            'priority': fixture.priority,
        } for fixture in provisioner.get_fixtures().to_list()]

        json.dumps(list, indent=2)

    def prepare(self, provisioner: str = ''):
        """ Run provisioner prepare """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.prepare()

    def apply(self, provisioner: str = ''):
        """ Run provisioner apply """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.apply()

    def check(self, provisioner: str = ''):
        """ Run provisioner check """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.check()

    def destroy(self, provisioner: str = ''):
        """ Run provisioner destroy """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.destroy()



"""

Dummy client plugin

"""

import logging
from typing import Dict, Any

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import ClientBase
from mirantis.testing.metta.fixtures import UCCTFixturesPlugin

logger = logging.getLogger('metta.contrib.dummy.client')


class DummyClientPlugin(ClientBase, UCCTFixturesPlugin):
    """ Dummy client class

    As with all dummies, this is a failsafe plugin, that should never throw any
    exceptions if used according to metta standards.

    It can be used as a placeholder during development, or it can be used to
    log client events and output for greater development and debugging.

    The client will log any method call, including unknown methods, and so it
    can be used in place of any client, if you don't need the methods to return
    anything
    """

    def __init__(self, environment: Environment, instance_id: str,
                 fixtures: Dict[str, Dict[str, Any]] = {}):
        """ Run the super constructor but also set class properties

        Overrides the ClientBase.__init__

        Arguments:
        ----------

        environment (metta.environment.Environment) : Environment in which this
            plugin exists.

        instance_id (str) : unique identifier for this plugin instance.

        fixtures (dict) : You can pass in some fixture definitions which this
            class will turn into fixtures and make retrievable.  This is a big
            part of the dummy.

        """
        ClientBase.__init__(self, environment, instance_id)

        fixtures = environment.add_fixtures_from_dict(plugin_list=fixtures)
        """ All fixtures added to this dummy plugin. """
        UCCTFixturesPlugin.__init__(self, fixtures)



import logging

from configerus.config import Config

from .plugin import METTAPlugin, Type

logger = logging.getLogger('metta.client')

METTA_PLUGIN_TYPE_CLIENT = Type.CLIENT
""" Fast access to the client plugin_id """

METTA_CLIENT_CONFIG_CLIENTS_LABEL = 'clients'
""" A centralized configerus label for multiple clients """
METTA_CLIENT_CONFIG_CLIENT_LABEL = 'client'
""" A centralized configerus label for a client """
METTA_CLIENT_CONFIG_CLIENTS_KEY = 'clients'
""" A centralized configerus key for multiple clients """
METTA_CLIENT_CONFIG_CLIENT_KEY = 'client'
""" A centralized configerus key for one client """


class ClientBase(METTAPlugin):
    """ Base class for client plugins """
    pass



import logging

from configerus.config import Config

from .plugin import METTAPlugin, Type

logger = logging.getLogger('metta.output')

METTA_PLUGIN_TYPE_OUTPUT = Type.OUTPUT
""" Fast access to the output plugin type """

METTA_OUTPUT_CONFIG_OUTPUTS_LABEL = 'outputs'
""" A centralized configerus load label for multiple outputs """
METTA_OUTPUT_CONFIG_OUTPUT_LABEL = 'output'
""" A centralized configerus load label for an output """
METTA_OUTPUT_CONFIG_OUTPUTS_KEY = 'outputs'
""" A centralized configerus key for multiple outputs """
METTA_OUTPUT_CONFIG_OUTPUT_KEY = 'output'
""" A centralized configerus key for one output """


class OutputBase(METTAPlugin):
    """ Base class for output plugins """
    pass

    def has_outout(self):
        """ Does this plugin have output """
        raise NotImplemented("this functionality has not yet been written.")



"""

A provisioner that combines multiple provisioners

"""
import logging
from typing import Dict, List, Any

from configerus.loaded import LOADED_KEY_ROOT
from mirantis.testing.metta.plugin import METTAPlugin, Type
from mirantis.testing.metta.fixtures import Fixtures, Fixture, UCCTFixturesPlugin
from mirantis.testing.metta.provisioner import ProvisionerBase

logger = logging.getLogger('metta.contrib.provisioner:combo')

COMBO_PROVISIONER_CONFIG_LABEL = 'provisioner'
""" Configerus label for loading config to set up this provisioner plugin """
COMBO_PROVISIONER_CONFIG_BACKENDS_KEY = 'backends'
""" Config key for backends list """


class ComboProvisionerPlugin(ProvisionerBase, UCCTFixturesPlugin):
    "Combo Provisioner plugin class"

    def __init__(self, environment, instance_id,
                 label: str = COMBO_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties

        Interpret provided config and configure the object with all of the needed
        pieces for executing terraform commands

        """
        ProvisionerBase.__init__(self, environment, instance_id)
        UCCTFixturesPlugin.__init__(self)

        logger.info("Preparing Combo provisioner")

        try:
            self.combo_config = self.environment.config.load(label)
        except KeyError as e:
            raise ValueError(
                "Combo provisioner could not find any configurations: {}".format(e)) from e

        self.backends = self.combo_config.get(
            [base, COMBO_PROVISIONER_CONFIG_BACKENDS_KEY])
        if not isinstance(self.backends, list):
            raise ValueError(
                "Combo provisioner could not understand the backend list.")

        for backend in self.backends:
            backend_instance_id = backend['instance_id']
            try:
                fixture = self.environment.fixtures.get_fixture(
                    type=Type.PROVISIONER, instance_id=backend_instance_id)
            except KeyError as e:
                raise ValueError(
                    "Combo provisioner was given a backend provisioner key that it could not correlated with a configured fixture: {}".format(backend_instance_id))

            if hasattr(backend, "priority"):
                fixture.priority = backend['priority']

            self.fixtures.add_fixture(fixture)

    def _get_ordered_backend_fixtures(self, low_to_high: bool = False):
        """ helper to get the sorted backend fixtures from lowest priority to highest, reversed if requested """
        backend_fixtures = self.fixtures.get_fixtures(
            type=Type.PROVISIONER).to_list()
        if low_to_high:
            backend_fixtures.reverse()
        return backend_fixtures

    def info(self):
        """ return structured data about self. """

        backends_info = []
        # List backends in high->low priority as this shows the order of apply
        for fixture in self._get_ordered_backend_fixtures():
            backend_info = {
                'fixture': {
                    'type': fixture.type.value,
                    'instance_id': fixture.instance_id,
                    'plugin_id': fixture.plugin_id
                }
            }

            plugin = fixture.plugin
            if hasattr(plugin, 'info'):
                backend_info.update(plugin.info())

            backends_info.append(backend_info)

        return {
            'backends': backends_info
        }

    def prepare(self, label: str = '', base: str = ''):
        """ Prepare the provisioner to apply resources """
        for backend_fixture in self._get_ordered_backend_fixtures():
            logger.info(
                "--> running backend prepare: [High->Low] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.prepare()

    def apply(self):
        """ bring a cluster to the configured state """
        for backend_fixture in self._get_ordered_backend_fixtures():
            logger.info(
                "--> running backend apply: [High->Low] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.apply()

    def destroy(self):
        """ remove all resources created for the cluster """
        for backend_fixture in self._get_ordered_backend_fixtures(
                low_to_high=True):
            logger.info(
                "--> running backend destroy: [Low->High] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.destroy()

    """ Fixture management

    We duplicate the UCCTFixturesPlugin methods, despite using it as a parent,
    so that we can identify as that object, but because we need to allow all
    backends to participate in fixture definition in order of priorit.

    We of course have to override the any method which depends on our ordered
    backend retrievals of get_fixtures() so that it doesn't run the parent
    get_fixtures.

    """

    def get_fixtures(self, type: Type = None, instance_id: str = '',
                     plugin_id: str = '') -> Fixtures:
        """ retrieve any matching fixtures from any of the backends """
        matches = Fixtures()
        for backend_fixture in self._get_ordered_backend_fixtures():
            plugin = backend_fixture.plugin
            if hasattr(plugin, 'get_fixtures'):
                for match in plugin.get_fixtures(
                        type=type, plugin_id=plugin_id, instance_id=instance_id).to_list():
                    matches.add_fixture(match)
        return matches

    def get_fixture(self, type: Type = None, instance_id: str = '',
                    plugin_id: str = '', exception_if_missing: bool = True) -> Fixture:
        """ retrieve the first matching fixture fomr backend in high-to-low order """

        for backend_fixture in self._get_ordered_backend_fixtures():
            plugin = backend_fixture.plugin
            if hasattr(plugin, 'get_fixture'):
                fixture = plugin.get_fixture(
                    type=type,
                    plugin_id=plugin_id,
                    instance_id=instance_id,
                    exception_if_missing=exception_if_missing)
                if fixture is not None:
                    return fixture

        if exception_if_missing:
            raise KeyError("No matching fixture was found")

    def get_plugin(self, type: Type = None, plugin_id: str = '',
                   instance_id: str = '', exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixtures """
        logger.info(
            "{}:execute: get_plugin({})".format(
                self.instance_id,
                type.value))
        fixture = self.get_fixture(
            type=type, plugin_id=plugin_id, instance_id=instance_id, exception_if_missing=exception_if_missing)

        if fixture is not None:
            return fixture.plugin

    def get_provisioner(self, plugin_id: str = '', instance_id: str = '',
                        exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture provisioner """
        return self.get_plugin(type=Type.PROVISIONER, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_output(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture outputs """
        return self.get_plugin(type=Type.OUTPUT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_client(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture clients """
        return self.get_plugin(type=Type.CLIENT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_workload(self, plugin_id: str = '', instance_id: str = '',
                     exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture workloads """
        return self.get_plugin(type=Type.WORKLOAD, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)



"""

CLI Plugins base

@see ./cli for the cli implementation

"""
import logging

import mirantis.testing.metta
from mirantis.testing.metta.plugin import METTAPlugin, Type

logger = logging.getLogger('metta.cli')

METTA_PLUGIN_TYPE_CLI = Type.CLI
""" Fast access to the output plugin type """


class CliBase(METTAPlugin):
    """ Base class for cli plugins """
    pass

    def exec(self):
        """ this plugin can execute """
        raise NotImplemented("this functionality has not yet been written.")



"""

A set of plugin instances kept as a managed set, we call these fixtures

"""
import logging
from typing import Dict, List, Any

from .plugin import (METTAPlugin, Type, METTA_PLUGIN_CONFIG_KEY_PLUGINID,
                     METTA_PLUGIN_CONFIG_KEY_INSTANCEID, METTA_PLUGIN_CONFIG_KEY_TYPE,
                     METTA_PLUGIN_CONFIG_KEY_PRIORITY)

logger = logging.getLogger('metta.fixtures')

METTA_FIXTURES_CONFIG_FIXTURES_LABEL = 'fixtures'
""" A centralized configerus load label for multiple fixtures """
METTA_FIXTURES_CONFIG_FIXTURE_KEY = 'fixture'
""" Config .get() key for a single fixture """

METTA_FIXTURE_VALIDATION_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        METTA_PLUGIN_CONFIG_KEY_PLUGINID: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_INSTANCEID: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_TYPE: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_PRIORITY: {
            'type': 'integer', 'minimum': 1, 'maximum': 100}
    },
    'required': [METTA_PLUGIN_CONFIG_KEY_PLUGINID]
}
""" json schema validation definition for a plugin """


class Fixture:
    """ A plugin wrapper struct that keep metadata about the plugin in a set """

    def __init__(self, plugin: object, type: Type,
                 plugin_id: str, instance_id: str, priority: int):
        """

        Parameters:
        -----------

        fixture : the fixture plugin instance

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        """
        self.type = type
        self.plugin_id = plugin_id
        self.instance_id = instance_id
        self.priority = priority
        self.plugin = plugin


class Fixtures:
    """ A set if plugins as a managed set

    A set of plugins that can be added in an arbitrary order but retrieved
    using filters and sorting.

    """

    def __init__(self):
        self.fixtures = []

    def __len__(self) -> int:
        """ Return how many plugin instances we have """
        return self.count()

    def __getitem__(self, instance_id: str) -> object:
        """ Handle subscription request

        For subscriptions assume that an instance_id is being retrieved and that
        a plugin is desired for return.

        Parameters:
        -----------

        instance_id (str) : Instance instance_id to look for

        Returns:

        Plugin object for highest priority plugin with the matching instance_id

        Raises:
        -------

        KeyError if the key cannot be matched,

        """
        return self.get_plugin(instance_id=instance_id)

    def merge_fixtures(self, merge_from: 'Fixtures'):
        """ merge fixture instances from another Fixtures object into this one

        Parameters:
        -----------

        merge_from (Fixtures) : fixture instance source

        """
        self.fixtures += merge_from.fixtures

    def new_fixture(self, plugin: object, type: Type,
                    plugin_id: str, instance_id: str, priority: int):
        """ Add a new fixture by providing the plugin instance and the metadata

        Create a new Fixture from the passed arguments and add it to the Fixtures set

        Parameters:
        -----------

        plugin : the fixture plugin instance

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        """
        fixture = Fixture(
            type=type,
            plugin_id=plugin_id,
            instance_id=instance_id,
            priority=priority,
            plugin=plugin)
        self.fixtures.append(fixture)
        return fixture

    def add_fixture(self, fixture: Fixture):
        """ Add an existing fixture

        Parameters:
        -----------

        fixture (Fixture) : existing fixture to add

        """
        self.fixtures.append(fixture)
        return fixture

    def to_list(self):
        """ retrieve this fixtures as a list """
        return sort_instance_list(self.fixtures)

    def count(self, type: Type = None, plugin_id: str = '',
              instance_id: str = ''):
        """ retrieve the first matching fixture object based on filters and priority

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        The fixture plugin object the sorted plugins that match the passed parameters,

        Raises:
        -------

        KeyError if exception_if_missing is True and no matching fixture was found

        """
        return len(self._filter_instances(
            type=type, plugin_id=plugin_id, instance_id=instance_id))

    def get_plugin(self, type: Type = None, plugin_id: str = '',
                   instance_id: str = '', exception_if_missing: bool = True):
        """ retrieve the first matching fixture object based on filters and priority

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------

        KeyError if exception_if_missing is True and no matching fixture was found

        """
        instance = self.get_fixture(
            type=type,
            plugin_id=plugin_id,
            instance_id=instance_id,
            exception_if_missing=exception_if_missing)

        if not instance is None:
            return instance.plugin

    def get_fixture(self, type: Type = None, plugin_id: str = '',
                    instance_id: str = '', exception_if_missing: bool = True) -> 'Fixture':
        """ retrieve the first matching fixture object based on filters and priority

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------

        KeyError if exception_if_missing is True and no matching fixture was found

        """
        instances = self.get_fixtures(
            type=type,
            plugin_id=plugin_id,
            instance_id=instance_id).to_list()

        if len(instances):
            return instances[0]
        if exception_if_missing:
            raise KeyError(
                "Could not find any matching fixture instances [type:{type}][plugin_id:{plugin_id}][instance_id:{instance_id}]".format(
                    type=type.value if not type is None else '',
                    plugin_id=plugin_id,
                    instance_id=instance_id))
        return None

    def get_plugins(self, type: Type = None, plugin_id: str = '',
                    instance_id: str = '') -> List[object]:
        """ retrieve the first matching fixture object based on filters and priority

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        A List of sorted Fixture fixture objects that match the arguments,
        possibly empty.

        """
        instances = self.get_fixtures(
            type=type,
            plugin_id=plugin_id,
            instance_id=instance_id).to_list()
        return [instance.plugin for instance in instances]

    def get_fixtures(self, type: Type = None, plugin_id: str = '',
                     instance_id: str = '') -> 'Fixtures':
        """ Retrieve an ordered filtered list of Fixtures

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        A List of Fixture structs that matched the arguments,
        possibly empty.
        If you want a sorted list see .to_list()

        """
        matches = Fixtures()
        [matches.add_fixture(match) for match in self._filter_instances(
            type=type, plugin_id=plugin_id, instance_id=instance_id)]
        return matches

    def get_filtered(self, type: Type = None,
                     plugin_id: str = '', instance_id: str = '') -> 'Fixtures':
        """ Get a new Fixtures object which is a filtered subset of this one """
        filtered = Fixtures()
        for instance in self._filter_instances(
                type=type, plugin_id=plugin_id, instance_id=instance_id):
            filtered.add_fixture(instance)
        return filtered

    def _filter_instances(self, type: Type = None,
                          plugin_id: str = '', instance_id: str = '') -> List[Fixture]:
        """ Filter the fixture instances down to a List

        Parameters:
        -----------

        Filtering parameters:

        type (.plugin.Type|str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------

        An unsorted List of Fixture structs that matched the arguments,
        possibly empty.

        Raises:
        -------

        KeyError if exception_if_missing is True and no matching fixture was found

        """
        matched_instances = []
        for plugin_instance in self.fixtures:
            # Could have one-lined this into a lambda but it would be
            # unreadable

            if plugin_id and not plugin_instance.plugin_id == plugin_id:
                continue
            elif instance_id and not plugin_instance.instance_id == instance_id:
                continue
            elif type and not type == plugin_instance.type:
                continue

            matched_instances.append(plugin_instance)

        return matched_instances


def sort_instance_list(list: List[Fixture]) -> List[Fixture]:
    """ Order a list of objects with a priority value from highest to lowest """
    return sorted(list, key=lambda i: 1 / i.priority if i.priority else 0)


class UCCTFixturesPlugin:
    """ Mixin class for plugins that hold fixtures """

    def __init__(self, fixtures: Fixtures = None):
        """ A Plugin that holds fixtures

        """
        if fixtures is None:
            fixtures = Fixtures()
        self.fixtures = fixtures
        """ Hold plugin fixtures, so that a provisioner can add output/clients etc """

    def get_fixtures(self, type: Type = None, instance_id: str = '',
                     plugin_id: str = '') -> Fixtures:
        """ retrieve an fixture plugin from the plugin """
        return self.fixtures.get_fixtures(
            type=type, plugin_id=plugin_id, instance_id=instance_id)

    def get_fixture(self, type: Type = None, instance_id: str = '',
                    plugin_id: str = '', exception_if_missing: bool = True) -> Fixture:
        """ retrieve an fixture plugin from the plugin """
        return self.fixtures.get_fixture(
            type=type, plugin_id=plugin_id, instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_plugin(self, type: Type = None, plugin_id: str = '',
                   instance_id: str = '', exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixtures """
        logger.info(
            "{}:execute: get_plugin({})".format(
                self.instance_id,
                type.value))
        return self.fixtures.get_plugin(
            type=type, plugin_id=plugin_id, instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_provisioner(self, plugin_id: str = '', instance_id: str = '',
                        exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture provisioner """
        return self.get_plugin(type=Type.PROVISIONER, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_output(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture outputs """
        return self.get_plugin(type=Type.OUTPUT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_client(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture clients """
        return self.get_plugin(type=Type.CLIENT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_workload(self, plugin_id: str = '', instance_id: str = '',
                     exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture workloads """
        return self.get_plugin(type=Type.WORKLOAD, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)




import os
import logging
import yaml
import json
import datetime
import subprocess
import shutil
from typing import Dict, List, Any

logger = logging.getLogger('metta_launchpad:launchpad')

METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT = './launchpad.yml'
""" Launchpad config configuration file key """

METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT = '.'
""" Launchpad Client default working dir """
METTA_LAUNCHPADCLIENT_BIN_PATH = 'launchpad'
""" Launchpad bin exec for the subprocess """

METTA_USER_LAUNCHPAD_CLUSTER_PATH = os.path.expanduser(
    os.path.join('~', '.mirantis-launchpad', 'cluster'))
""" the str path to where launchpad keeps its user config """
METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH = 'bundle'
""" str path to user bundle config can be found when it is downloaded """
METTA_USER_LAUNCHPAD_BUNDLE_META_FILE = 'meta.json'
""" str filename for the meta file in the client bundle path """

LAUNCHPAD_CLIENT_BUNDLE_RETRY_COUNT_DEFAULT = 2
""" default value for how many times we should retry the client bundle download """


class LaunchpadClient:
    """ shell client for interacting with the launchpad bin """

    def __init__(self, config_file: str = METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
                 working_dir: str = METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
                 cluster_name_override: str = '', accept_license: bool = False,
                 disable_telemetry: bool = False, disable_upgrade_check: bool = True,
                 debug: bool = False):
        """
        Parameters:
        -----------

        config_file (str) : path to launchpad config file, typically
            launchpad.yml

        working_dir (str) : full config file path.
            this typically plays a role in interpreting file paths from the
            config for things like ssh keys.  The client will use that path for
            python subprocess execution

        cluster_name_override (str) : string name for the cluster which should
            be used instead of trying to read a value from config.  This can
            can help in some scenarios where config cannot be read outside of the
            client, and so cluster name is hard to discover.

        accept_license (bool) : passed to the launchpad client to tell it to
            accept the license on first use.

        disable_telemetry (bool) : passed to the launchpad client to tell it to
            disable telemetry on actions taken.

        disable_upgrade_check (bool) : passed to the launchpad client to tell it
            to disable checking to see if a new client version is available.

        debug (bool) : passed to the launchpad client to tell it to enable
            verbose debugging output.

        """
        self.config_file = config_file
        """ Path to config file """
        self.working_dir = working_dir
        """ Python subprocess working dir to execute launchpad in.
            This may be relevant in cases where ssh keys have a relative path """

        if cluster_name_override:
            self.cluster_name_override = str(cluster_name_override)
            """ if not empty this will be used as a cluster name instead of taking
                it from the yaml file """
        else:
            self.cluster_name_override = None

        self.bin = METTA_LAUNCHPADCLIENT_BIN_PATH
        """ shell execution target for launchpad """

        self.client_bundle_retry_count = LAUNCHPAD_CLIENT_BUNDLE_RETRY_COUNT_DEFAULT
        """ how many times to rety a client bundle download """

        self.disable_telemetry = disable_telemetry
        self.accept_license = accept_license
        self.disable_upgrade_check = disable_upgrade_check

        self.debug = debug
        """ should launchpad be run with verbose output enabled ? """

    def version(self):
        """ Output launchpad client version """
        self._run(['version'])

    def apply(self, debug: bool = False):
        """ Install using the launchpad client """
        self._run(['apply'], debug=debug)

    def exec(self, host_index: int, cmds: List[str]):
        """ execute a command on a host index """
        client_config = self.describe_config()
        hosts = client_config['spec']['hosts']
        host = hosts[host_index]
        if 'ssh' in host:
            target = host['ssh']['address']
        if 'winrm' in host:
            target = host['winrm']['address']

        args = ['exec', '--target', target]
        args.extend(cmds)
        self._run(args)

    def exec_interactive(self, host_index: int, cmds: List[str]):
        """ execute a command on a host index """
        client_config = self.describe_config()
        hosts = client_config['spec']['hosts']
        host = hosts[host_index]
        if 'ssh' in host:
            target = host['ssh']['address']
        if 'winrm' in host:
            target = host['winrm']['address']
        args = ['exec', '--interactive', '--target', target]
        args.extend(cmds)
        self._run(args)

    def reset(self, quick: bool = False):
        """ Uninstall using the launchpad client """
        if os.path.isfile(self.config_file):
            if not quick:
                self._run(['reset', '--force'])

            os.remove(self.config_file)
            self.rm_client_bundles()

    def register(self, name: str, email: str, company: str):
        """ Uninstall using the launchpad client """
        return self._run(['register', '--name', name,
                          '--email', email, '--company', company])

    def describe_config(self):
        """ Return the launchpad config report as umarshalled yaml """
        return yaml.safe_load(
            self._run(['describe', 'config'], return_output=True))

    def describe(self, report: str):
        """ Output one of the launchpad reports """
        self._run(['describe', report])

    def bundle_users(self):
        """ list bundle users which have been downloaded """
        return self._mke_client_downloaded_bundle_user_paths().keys()

    def bundle(self, user: str, reload: bool = False):
        """ Retrieve a client bundle and return the metadata as a dict """
        client_bundle_path = self._mke_client_bundle_path(user)
        client_bundle_meta_file = os.path.join(
            client_bundle_path, METTA_USER_LAUNCHPAD_BUNDLE_META_FILE)

        if reload or not os.path.isfile(client_bundle_meta_file):

            # @NOTE currently client bundle downloads are flaky.  They fail about 1/5 times
            #    with unclear TLS issues.  Because the failures are intermittent, we should
            #    just try again

            for i in range(1, self.client_bundle_retry_count):
                try:
                    self._run(["client-config", user])
                    break
                except Exception as e:
                    logger.warn(
                        "Attempt {} to download bundle.  Assuming flaky behaviour and trying again : {}".format(
                            i, e))

            else:
                raise Exception(
                    "Numerous attempts to download the client bundle have failed.")

        data = {}
        """ Will hold data pulled from the client meta data file """
        try:
            with open(client_bundle_meta_file) as json_file:
                data = json.load(json_file)

                # helm complains if this file has loose permissions
                client_bundle_kubeconfig_file = os.path.join(
                    client_bundle_path, 'kube.yml')
                os.chmod(client_bundle_kubeconfig_file, 0o600)
        except FileNotFoundError as e:
            raise ValueError(
                "failed to open the launchpad client bundle meta file : {}".format(client_bundle_meta_file)) from e

        # Not sure why this isn't in there:
        data['Endpoints']['kubernetes']['kubeconfig'] = client_bundle_kubeconfig_file
        # add some stuff that a client bundle always has
        data['path'] = client_bundle_path
        data['modified'] = datetime.datetime.fromtimestamp(
            os.path.getmtime(client_bundle_meta_file)).strftime("%Y-%m-%d %H:%M:%S")
        # this stuff should already be in the bundle, but it isn't
        data['tls_paths'] = {
            'docker': os.path.join(client_bundle_path, 'tls', 'docker'),
            'kubernetes': os.path.join(client_bundle_path, 'tls', 'kubernetes'),
        }

        return data

    def rm_client_bundles(self):
        """ remove any downloaded client bundles """
        try:
            base = os.path.join(
                METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                self._cluster_name(),
                METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH)
        except BaseException:
            # most likely we could't determine a cluster_name, because we don't
            # have a config file.
            return

        if os.path.isdir(base):
            shutil.rmtree(base)

    def _mke_client_bundle_root(self):
        """  root path to the launchpad user conf """
        return os.path.join(
            METTA_USER_LAUNCHPAD_CLUSTER_PATH, self._cluster_name())

    def _mke_client_bundle_path(self, user: str):
        """ find the path to a client bundle for a user whether or not it has been downloaded """
        return os.path.join(METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                            self._cluster_name(), METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH, user)

    def _mke_client_downloaded_bundle_user_paths(self) -> Dict[str, str]:
        """ return a map of user names to downloaded bundle paths """
        try:
            base = os.path.join(
                METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                self._cluster_name(),
                METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH)
            return {userdir: os.path.join(base, userdir) for userdir in os.listdir(
                base) if os.path.isdir(os.path.join(base, userdir))}
        except (ValueError, FileNotFoundError) as e:
            logger.debug(
                "Could not get user bundles names as there is no launchpad targets to check against.")
            return {}

    def _cluster_name(self):
        """ get the cluster name from the config file

        @TODO we should cache it, but then do we need to invalidate cache?

        """

        if self.cluster_name_override is not None:
            return str(self.cluster_name_override)

        try:
            with open(self.config_file) as config_file_object:
                self.config_data = yaml.load(
                    config_file_object, Loader=yaml.FullLoader)
                """ keep a parsed copy of the launchpad file """
        except FileNotFoundError as e:
            raise ValueError(
                "Launchpad yaml file could not be opened: {}".format(
                    self.config_file)) from e
        except Exception as e:
            raise ValueError(
                "Launchpad yaml file had unexpected contents: {} : {}".format(
                    self.config_file, e)) from e

        if not isinstance(self.config_data, dict):
            raise ValueError(
                "Launchpad yaml file had unexpected contents: {}".format(self.config_file))

        try:
            return str(self.config_data['metadata']['name'])
        except KeyError:
            raise ValueError(
                'Launchpad yaml file did not container a cluster name')

    def _run(self, args: List[str] = ['help'],
             return_output=False, debug: bool = False):
        """ Run a launchpad command

        Parameters:

        args (List[str]) : arguments to pass to the launchpad bin using
            subprocess

        return_output (bool) : this call should capture the exec output and
            return it instead of sending it to stdout

        debug (bool) : override class debug value if True

        """

        """ if the command passed uses a config file, add the flag for it """
        if not args[0] in ['help', 'version']:
            args = [args[0]] + ['-c', self.config_file] + args[1:]

        cmd = [self.bin]

        cmd += [args[0]]

        if debug or self.debug:
            cmd += ['--debug']
        if self.disable_telemetry:
            cmd += ['--disable-telemetry']
        if self.accept_license:
            cmd += ['--accept-license']
        if self.disable_upgrade_check:
            cmd += ['--disable-upgrade-check']

        if len(args) > 1:
            cmd += args[1:]

        if return_output:
            logger.debug(
                "running launchpad command with output capture: %s",
                " ".join(cmd))
            exec = subprocess.run(
                cmd,
                cwd=self.working_dir,
                shell=False,
                stdout=subprocess.PIPE)
            exec.check_returncode()
            return exec.stdout.decode('utf-8')
        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            exec = subprocess.run(
                cmd, cwd=self.working_dir, check=True, text=True)
            exec.check_returncode()



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

from .output import OutputGroup

logger = logging.getLogger('metta.cli.provisioner')


class ProvisionerCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'provisioner': ProvisionerGroup(self.environment)
        }


class ProvisionerGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def list(self, raw: bool = False):
        """ List all provisioners """
        list = [
            plugin.instance_id for plugin in self.environment.fixtures.get_plugins(
                type=Type.PROVISIONER)]

        if raw:
            return list
        else:
            return json.dumps(list)

    def _select_provisioner(self, instance_id: str = ''):
        """ Pick a matching provisioner """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, instance_id=instance_id)
        else:
            # Get the highest priority provisioner
            return self.environment.fixtures.get_fixture(type=Type.PROVISIONER)

    def info(self, provisioner: str = '', deep: bool = True):
        """ get info about a provisioner plugin """
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            if hasattr(fixture.plugin, 'get_fixtures'):
                fixtures = {}
                for sub_fixture in fixture.plugin.get_fixtures().to_list():
                    fixture_info = {
                        'fixture': {
                            'type': sub_fixture.type.value,
                            'plugin_id': sub_fixture.plugin_id,
                            'instance_id': sub_fixture.instance_id,
                            'priority': sub_fixture.priority,
                        }
                    }
                    if hasattr(sub_fixture.plugin, 'info'):
                        plugin_info = sub_fixture.plugin.info()
                        if isinstance(plugin_info, dict):
                            fixture_info.update(plugin_info)
                    fixtures[sub_fixture.instance_id] = fixture_info
                info['fixtures'] = fixtures

        return json.dumps(info, indent=2, default=lambda X: "{}".format(X))

    def fixtures(self, provisioner: str = '', deep: bool = False):
        """ List all fixtures for this provisioner """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')

        list = {}
        for fixture in provisioner.get_fixtures().to_list():

            info = {
                'fixture': {
                    'type': fixture.type.value,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }

            if deep:
                if hasattr(fixture.plugin, 'info'):
                    plugin_info = fixture.plugin.info()
                    if isinstance(plugin_info, dict):
                        info.update(plugin_info)

                if hasattr(fixture.plugin, 'get_fixtures'):
                    fixtures = {}
                    for sub_fixture in fixture.plugin.get_fixtures().to_list():
                        fixture_info = {
                            'fixture': {
                                'type': sub_fixture.type.value,
                                'plugin_id': sub_fixture.plugin_id,
                                'instance_id': sub_fixture.instance_id,
                                'priority': sub_fixture.priority,
                            }
                        }
                        if hasattr(sub_fixture.plugin, 'info'):
                            plugin_info = sub_fixture.plugin.info()
                            if isinstance(plugin_info, dict):
                                fixture_info.update(plugin_info)
                        fixtures[sub_fixture.instance_id] = fixture_info
                    info['fixtures'] = fixtures

            list[fixture.instance_id] = info

        return json.dumps(list, indent=2)

    def prepare(self, provisioner: str = ''):
        """ Run provisioner prepare """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.prepare()

    def apply(self, provisioner: str = ''):
        """ Run provisioner apply """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.apply()

    def destroy(self, provisioner: str = ''):
        """ Run provisioner destroy """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.destroy()



"""

Provisioning

"""

import logging

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT

from .plugin import METTAPlugin, Type
from .fixtures import Fixtures

logger = logging.getLogger('metta.provisioner')

METTA_PLUGIN_TYPE_PROVISIONER = Type.PROVISIONER
""" Fast access to the Provisioner plugin_id """

METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL = 'provisioners'
""" A centralized configerus load labe for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL = 'provisioner'
""" A centralized configerus load label a provisioner """
METTA_PROVISIONER_CONFIG_PROVISIONERS_KEY = 'provisioners'
""" A centralized configerus key for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_KEY = 'provisioner'
""" A centralized configerus key for one provisioner """


class ProvisionerBase(METTAPlugin):
    "Base Provisioner plugin class"

    def prepare(self, label: str = '', base: str = ''):
        """ Prepare the provisioner to apply resources

        Initial Provisioner plugin is expected to be of very low cost until
        prepare() is executed.  At this point the plugin should load any config
        and perform any validations needed.
        The plugin should not create any resources but it is understood that
        there may be a cost of preparation.

        Provisioners are expected to load a lot of config to self-program.
        Because of this, they allow passing of a configerus label for .load()
        and a base for .get() in case there is an alterante config source
        desired.

        """
        raise NotImplementedError(
            'This provisioner has not yet implemented prepare')

    def apply(self):
        """ bring a cluster to the configured state """
        raise NotImplementedError(
            'This provisioner has not yet implemented apply')

    def destroy(self):
        """ remove all resources created for the cluster """
        raise NotImplementedError(
            'This provisioner has not yet implemented destroy')



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta import environment_names, get_environment
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

logger = logging.getLogger('metta.cli.environment')


class EnvironmentCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'environment': EnvironmentGroup(self.environment)
        }


class EnvironmentGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def list(self, raw: bool = False):
        """ List all of the environment names  """
        names = environment_names()
        if raw:
            return names
        else:
            return json.dumps(names)

    def _environment(self, environment: str = ''):
        """ select an environment """
        if not environment:
            return self.environment
        return get_environment(environment)

    def name(self, environment: str = ''):
        """ return env name """
        environment = self._environment(environment)
        return environment.name

    def info(self, environment: str = ''):
        """ return info about an environment """
        environment = self._environment(environment)

        info = {
            'name': environment.name
        }

        if len(environment.states) > 0:
            info['states'] = {
                'available': environment.states,
                'active': environment.state
            }

        return json.dumps(info, indent=2, default=lambda X: "{}".format(X))

    def bootstraps(self, environment: str = ''):
        """ List bootstraps that have been applied to the environment """
        environment = self._environment(environment)

        list = [bootstrap for bootstrap in environment.bootstrapped]

        return json.dumps(list, indent=2, default=lambda X: "{}".format(X))



"""

Terraform METTA provisioner plugin

"""

import logging
import os
from typing import List, Any

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.plugin import METTAPlugin, Type
from mirantis.testing.metta.fixtures import Fixtures, UCCTFixturesPlugin, METTA_FIXTURES_CONFIG_FIXTURES_LABEL
from mirantis.testing.metta.provisioner import ProvisionerBase
from mirantis.testing.metta.output import OutputBase
from mirantis.testing.metta_common import METTA_PLUGIN_ID_OUTPUT_DICT, METTA_PLUGIN_ID_OUTPUT_TEXT

from .terraform import TerraformClient

logger = logging.getLogger('metta_terraform:provisioner')

METTA_TERRAFORM_PROVISIONER_PLUGIN_ID = 'metta_terraform'
""" Terraform provisioner plugin id """

TERRAFORM_PROVISIONER_CONFIG_LABEL = 'terraform'
""" config label loading the terraform config """
TERRAFORM_PROVISIONER_CONFIG_ROOT_PATH_KEY = 'root.path'
""" config key for a base path that should be used for any relative paths """
TERRAFORM_PROVISIONER_CONFIG_PLAN_PATH_KEY = 'plan.path'
""" config key for the terraform plan path """
TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY = 'state.path'
""" config key for the terraform state path """
TERRAFORM_PROVISIONER_CONFIG_VARS_KEY = 'vars'
""" config key for the terraform vars Dict, which will be written to a file """
TERRAFORM_PROVISIONER_CONFIG_VARS_PATH_KEY = 'vars_path'
""" config key for the terraform vars file path, where the plugin will write to """
TERRAFORM_PROVISIONER_DEFAULT_VARS_FILE = 'metta_terraform.tfvars.json'
""" Default vars file if none was specified """
TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH = 'metta-state'
""" Default vars file if none was specified """

TERRAFORM_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'type': {'type': 'string'},
        'plugin_id': {'type': 'string'},

        'root': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            }
        },
        'plan': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            },
            'required': ['path']
        },
        'state': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            }
        },
        'vars_path': {
            'type': 'string'
        },
        'vars': {
            'type': 'object'
        }
    }
}
""" Validation jsonschema for terraform config contents """
TERRAFORM_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: TERRAFORM_VALIDATE_JSONSCHEMA}
""" configerus validation target to match the above config """


class TerraformProvisionerPlugin(ProvisionerBase, UCCTFixturesPlugin):
    """ Terraform provisioner plugin

    Provisioner plugin that allows control of and interaction with a terraform
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a terraform binary, so you have to install
       terraform in the environment

    ## Usage

    ### Plan

    The plan must exists somewhere on disk, and be accessible.

    You must specify the path and related configuration in config, which are read
    in the .prepare() execution.

    ### Vars/State

    This plugin reads TF vars from config and writes them to a vars file.  We
    could run without relying on vars file, but having a vars file allows cli
    interaction with the cluster if this plugin messes up.

    You can override where Terraform vars/state files are written to allow sharing
    of a plan across test suites.

    """

    def __init__(self, environment, instance_id,
                 label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties

        Interpret provided config and configure the object with all of the needed
        pieces for executing terraform commands

        """
        super(ProvisionerBase, self).__init__(environment, instance_id)

        logger.info("Preparing Terraform setting")

        self.terraform_config_label = label
        """ configerus load label that should contain all of the config """
        self.terraform_config_base = base
        """ configerus get key that should contain all tf config """

        self.terraform_config = self.environment.config.load(
            self.terraform_config_label)
        """ get a configerus LoadedConfig for the terraform label """

        # Run confgerus validation on the config using our above defined
        # jsonschema
        try:
            self.terraform_config.get(
                base, validator=TERRAFORM_VALIDATE_TARGET)
        except ValidationError as e:
            raise ValueError(
                "Terraform config failed validation: {}".format(e)) from e

        fixtures = self.environment.add_fixtures_from_config(
            label=self.terraform_config_label,
            base=[self.terraform_config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL])
        """ All fixtures added to this provisioner plugin. """
        UCCTFixturesPlugin.__init__(self, fixtures)

        self.root_path = self.terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_ROOT_PATH_KEY], default='')
        """ all relative paths will have this joined as their base """

        try:
            self.working_dir = self.terraform_config.get(
                [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_PLAN_PATH_KEY])
            """ all subprocess commands for terraform will be run in this path """
        except Exception as e:
            raise ValueError(
                "Plugin config did not give us a working/plan path: {}".format(self.terraform_config.data)) from e
        if not os.path.isabs(self.working_dir):
            if self.root_path:
                self.working_dir = os.path.join(
                    self.root_path, self.working_dir)
            self.working_dir = os.path.abspath(self.working_dir)

        state_path = self.terraform_config.get(
            [
                self.terraform_config_base,
                TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY],
            default=os.path.join(
                self.working_dir,
                TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH))
        """ terraform state path """
        if not os.path.isabs(state_path):
            if self.root_path:
                state_path = os.path.join(self.root_path, state_path)
            state_path = os.path.abspath(state_path)

        self.vars = self.terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_VARS_KEY], default={})
        """ List of vars to pass to terraform.  Will be written to a file """

        vars_path = self.terraform_config.get(
            [
                self.terraform_config_base,
                TERRAFORM_PROVISIONER_CONFIG_VARS_PATH_KEY],
            default=os.path.join(
                self.working_dir,
                TERRAFORM_PROVISIONER_DEFAULT_VARS_FILE))
        """ vars file containing vars which will be written before running terraform """
        if not os.path.isabs(vars_path):
            if self.root_path:
                vars_path = os.path.join(self.root_path, vars_path)
            vars_path = os.path.abspath(vars_path)

        logger.info("Creating Terraform client")

        self.tf = TerraformClient(
            working_dir=os.path.realpath(self.working_dir),
            state_path=os.path.realpath(state_path),
            vars_path=os.path.realpath(vars_path),
            variables=self.vars)
        """ TerraformClient instance """

        # if the cluster is already provisioned then we can get outputs from it
        try:
            self._get_outputs_from_tf()
        except Exception:
            pass

    def info(self):
        """ get info about a provisioner plugin """
        plugin = self
        client = plugin.tf

        info = {}
        info['plugin'] = {
            'terraform_config_label': plugin.terraform_config_label,
            'terraform_config_base': plugin.terraform_config_base
        },
        info['client'] = {
            'vars': client.vars,
            'working_dir': client.working_dir,
            'state_path': client.state_path,
            'vars_path': client.vars_path,
            'terraform_bin': client.terraform_bin
        }

        fixtures = {}
        for fixture in self.get_fixtures().to_list():
            fixture_info = {
                'fixture': {
                    'type': fixture.type.value,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    fixture_info.update(plugin_info)
            fixtures[fixture.instance_id] = plugin_info
        info['fixtures'] = fixtures

        info['helper'] = {
            'commands': {
                'init': "{bin} -chdir={working_dir} init".format(bin=client.terraform_bin, working_dir=client.working_dir),
                'plan': "{bin} -chdir={working_dir} plan -var-file={vars_path} -state={state_path}".format(bin=client.terraform_bin, working_dir=client.working_dir, vars_path=client.vars_path, state_path=client.state_path),
                'apply': "{bin} -chdir={working_dir} apply -var-file={vars_path} -state={state_path}".format(bin=client.terraform_bin, working_dir=client.working_dir, vars_path=client.vars_path, state_path=client.state_path),
                'destroy': "{bin} -chdir={working_dir} destroy -var-file={vars_path} -state={state_path}".format(bin=client.terraform_bin, working_dir=client.working_dir, vars_path=client.vars_path, state_path=client.state_path),
                'output': "{bin} -chdir={working_dir} output -state={state_path}".format(bin=client.terraform_bin, working_dir=client.working_dir, state_path=client.state_path)
            }
        }

        return info

    def prepare(self):
        """ run terraform init """
        logger.info("Running Terraform INIT")
        self.tf.init()

    def check(self):
        """ Check that the terraform plan is valid """
        logger.info("Running Terraform PLAN")
        self.tf.plan()

    def apply(self):
        """ Create all terraform resources described in the plan """
        logger.info("Running Terraform APPLY")
        self.tf.apply()
        self._get_outputs_from_tf()

    def destroy(self):
        """ Remove all terraform resources in state """
        logger.info("Running Terraform DESTROY")
        self.tf.destroy()
        self.fixtures = Fixtures()

    def clean(self):
        """ Remove terraform run resources from the plan """
        logger.info("Running Terraform CLEAN")
        dot_terraform = os.path.join(self.working_dir, '.terraform')
        if os.isdir(dot_terraform):
            shutil.rmtree(dot_terraform)

    """ Cluster Interaction """

    def _get_outputs_from_tf(self) -> Fixtures:
        """ retrieve an output from terraform

        For other METTA plugins we can just load configuration, and creating
        output plugin instances from various value in config.

        We do that here, but  we also want to check of any outputs exported by
        the terraform root module, which we get using the tf client.

        If we find a root module output without a matching config output
        defintition then we make some assumptions about plugin type and add it
        to the list. We make some simple investigation into output plugin types
        and pick either the contrib.common.dict or contrib.common.text plugins.

        If we find a root module output that matches an output that was declared
        in config then we use that.  This allows config to define a plugin_id
        which will then be populated automatically.  If you know what type of
        data you are expecting from a particular tf output then you can prepare
        and config for it to do things like setting default values.

        Priorities can be used in the config.

        Returns:
        --------

        A Fixtures set of plugins.

        """

        # now we ask TF what output it nows about and merge together those as
        # new output plugins.
        # tf.outputs() produces a list of (sensitive:bool, type: [str,  object,
        # value:Any])
        for output_key, output_struct in self.tf.output().items():
            # we only know how to create 2 kinds of outputs
            output_sensitive = bool(output_struct['sensitive'])
            """ Whether or not the output contains sensitive data """
            output_type = output_struct['type'][0]
            """ String output primitive type (usually string|object|number) """
            output_spec = output_struct['type'][1]
            """ A structured spec for the type """
            output_value = output_struct['value']
            """ output value """

            # see if we already have an output plugin for this name
            fixture = self.fixtures.get_fixture(
                type=Type.OUTPUT,
                instance_id=output_key,
                exception_if_missing=False)
            if not fixture:
                if output_type == 'object':
                    fixture = self.environment.add_fixture(
                        type=Type.OUTPUT,
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_DICT,
                        instance_id=output_key,
                        priority=self.environment.plugin_priority(delta=5),
                        arguments={'data': output_value})
                else:
                    fixture = self.environment.add_fixture(
                        type=Type.OUTPUT,
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_TEXT,
                        instance_id=output_key,
                        priority=self.environment.plugin_priority(delta=5),
                        arguments={'text': str(output_value)})

                self.fixtures.add_fixture(fixture)
            else:
                if hasattr(fixture.plugin, 'set_data'):
                    fixture.plugin.set_data(output_value)
                elif hasattr(fixture.plugin, 'set_text'):
                    fixture.plugin.set_text(str(output_value))



import logging

from configerus.config import Config

from .plugin import METTAPlugin, Type
from .fixtures import Fixtures

logger = logging.getLogger('metta.workload')

METTA_PLUGIN_ID_WORKLOAD = Type.WORKLOAD
""" Fast access to the workload type """

METTA_WORKLOAD_CONFIG_WORKLOADS_LABEL = 'workloads'
""" A centralized configerus load label for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL = 'workload'
""" A centralized configerus load label for a workload """
METTA_WORKLOAD_CONFIG_WORKLOADS_KEY = 'workloads'
""" A centralized configerus key for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_KEY = 'workload'
""" A centralized configerus key for one workload """


class WorkloadBase(METTAPlugin):
    """ Base class for workload plugins """

    def create_instance(self, fixtures: Fixtures) -> 'WorkloadInstanceBase':
        """ createe a workload instance from a set of fixtures """


class WorkloadInstanceBase:
    """ An instance of a running workload """

    def apply(self):
        """ Run the workload

        @NOTE Needs a kubernetes client fixture to run.  Use .set_fixtures() first

        """
        raise NotImplementedError(
            "This workload plugin has not implemented apply()")

    def destroy(self):
        """ destroy any created resources """
        raise NotImplementedError(
            "This workload plugin has not implemented destroy()")
